import re
from contextlib import contextmanager
from datetime import datetime
from json.decoder import JSONDecodeError
from typing import Optional, Any, Callable, Iterator, cast

import httpx
import rpcq
from dateutil.parser import parse as parsedate
from dateutil.tz import tzutc
from qcs_api_client.client import QCSClientConfiguration, build_sync_client
from qcs_api_client.models import (
    EngagementWithCredentials,
    CreateEngagementRequest,
    EngagementCredentials,
)
from qcs_api_client.operations.sync import create_engagement
from qcs_api_client.types import Response
from rpcq import ClientAuthConfig

from pyquil.api._errors import ApiError, UnknownApiError, TooManyQubitsError, error_mapping
from pyquil.api._logger import logger


class Client:
    """
    Class for housing application configuration and interacting with network resources.
    """

    _http: Optional[httpx.Client]
    _config: QCSClientConfiguration
    _has_default_config: bool
    _engagement: Optional[EngagementWithCredentials] = None

    def __init__(
        self,
        *,
        http: Optional[httpx.Client] = None,
        configuration: Optional[QCSClientConfiguration] = None,
    ):
        """
        Instantiate a client.

        :param http: optional underlying HTTP client. If none is provided, a default
               client will be created using ``configuration``.
        :param configuration: QCS configuration.
        """
        self._config = configuration or QCSClientConfiguration.load()
        self._has_default_config = configuration is None
        self._http = http

    def post_json(self, url: str, json: Any, timeout: float = 5) -> httpx.Response:
        """
        Post JSON to a URL. Will raise an exception for response statuses >= 400.

        :param url: URL to post to.
        :param json: JSON body of request.
        :param timeout: Time limit for request, in seconds.
        :return: HTTP response corresponding to request.
        """
        logger.debug("Sending POST request to %s. Body: %s", url, json)
        with self._http_client() as http:  # type: httpx.Client
            res = http.post(url, json=json, timeout=timeout)
            if res.status_code >= 400:
                raise _parse_error(res)
        return res

    def qcs_request(self, request_fn: Callable[..., Response[Any]], **kwargs: Any) -> Any:
        """
        Execute a QCS request.

        :param request_fn: Request function (from ``qcs_api_client.operations.sync``).
        :param kwargs: Arguments to pass to request function.
        :return: HTTP response corresponding to request.
        """
        with self._http_client() as http:  # type: httpx.Client
            return request_fn(client=http, **kwargs).parsed

    def compiler_rpcq_request(
        self, method_name: str, *args: Any, timeout: Optional[float] = None, **kwargs: Any
    ) -> Any:
        """
        Execute a remote function against the Quil compiler.

        :param method_name: Method name.
        :param args: Arguments that will be passed to the remote function.
        :param timeout: Optional time limit for request, in seconds.
        :param kwargs: Keyword arguments that will be passed to the remote function.
        :return: Result from remote function.
        """
        return self._rpcq_request(self.quilc_url, method_name, *args, timeout=timeout, **kwargs)

    def processor_rpcq_request(
        self,
        quantum_processor_id: str,
        method_name: str,
        *args: Any,
        timeout: Optional[float] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Execute a remote function against a processor endpoint. If there is no current engagement,
        or if it is invalid, a new engagement will be requested for the given processor.

        :param quantum_processor_id: Processor to engage.
        :param method_name: Method name.
        :param args: Arguments that will be passed to the remote function.
        :param timeout: Optional time limit for request, in seconds.
        :param kwargs: Keyword arguments that will be passed to the remote function.
        :return: Result from remote function.
        """
        # TODO(andrew): handle multiple engagements at once (per processor)
        if not _engagement_valid(quantum_processor_id, self._engagement):
            self._engagement = self._create_engagement(quantum_processor_id)

        assert self._engagement is not None

        return self._rpcq_request(
            self._engagement.address,
            method_name,
            *args,
            timeout=timeout,
            auth_config=_to_auth_config(self._engagement.credentials),
            **kwargs,
        )

    def _rpcq_request(
        self,
        endpoint: str,
        method_name: str,
        *args: Any,
        timeout: Optional[float] = None,
        auth_config: Optional[ClientAuthConfig] = None,
        **kwargs: Any,
    ) -> Any:
        client = rpcq.Client(endpoint, auth_config=auth_config)
        response = client.call(method_name, *args, rpc_timeout=timeout, **kwargs)  # type: ignore
        client.close()  # type: ignore
        return response

    def reset(self) -> None:
        """
        Clears current engagement and reloads configuration (if not overridden in constructor).
        """
        self._engagement = None
        if self._has_default_config:
            self._config = QCSClientConfiguration.load()

    @property
    def qvm_url(self) -> str:
        """
        QVM URL from client configuration.
        """
        return self._config.profile.applications.pyquil.qvm_url

    @property
    def quilc_url(self) -> str:
        """
        Quil compiler URL from client configuration.
        """
        return self._config.profile.applications.pyquil.quilc_url

    def qvm_version(self) -> str:
        """
        Get QVM version string.
        """
        response = self.post_json(self.qvm_url, {"type": "version"})
        split_version_string = response.text.split()
        try:
            qvm_version = split_version_string[0]
        except ValueError:
            raise TypeError(f"Malformed version string returned by the QVM: {response.text}")
        return qvm_version

    @contextmanager
    def _http_client(self) -> Iterator[httpx.Client]:
        if self._http is None:
            with build_sync_client(configuration=self._config) as client:  # type: httpx.Client
                yield client
        else:
            yield self._http

    def _create_engagement(self, quantum_processor_id: str) -> EngagementWithCredentials:
        return cast(
            EngagementWithCredentials,
            self.qcs_request(
                create_engagement,
                json_body=CreateEngagementRequest(quantum_processor_id=quantum_processor_id),
            ),
        )


def _engagement_valid(quantum_processor_id: str, engagement: Optional[EngagementWithCredentials]) -> bool:
    if engagement is None:
        return False

    return all(
        [
            engagement.credentials.client_public != "",
            engagement.credentials.client_secret != "",
            engagement.credentials.server_public != "",
            parsedate(engagement.expires_at) > datetime.now(tzutc()),
            engagement.address != "",
            engagement.quantum_processor_id == quantum_processor_id,
        ]
    )


def _to_auth_config(credentials: EngagementCredentials) -> ClientAuthConfig:
    return rpcq.ClientAuthConfig(
        client_secret_key=credentials.client_secret.encode(),
        client_public_key=credentials.client_public.encode(),
        server_public_key=credentials.server_public.encode(),
    )


def _parse_error(res: httpx.Response) -> ApiError:
    """
    Errors should contain a "status" field with a human readable explanation of
    what went wrong as well as a "error_type" field indicating the kind of error that can be mapped
    to a Python type.

    There's a fallback error UnknownApiError for other types of exceptions (network issues, api
    gateway problems, etc.)
    """
    try:
        body = res.json()
    except JSONDecodeError:
        raise UnknownApiError(res.text)

    if "error_type" not in body:
        raise UnknownApiError(str(body))

    error_type = body["error_type"]
    status = body["status"]

    if re.search(r"[0-9]+ qubits were requested, but the QVM is limited to [0-9]+ qubits.", status):
        return TooManyQubitsError(status)

    error_cls = error_mapping.get(error_type, UnknownApiError)
    return error_cls(status)



import numpy as np

from pyquil import Program
from pyquil.experiment._program import (
    measure_qubits,
    parameterized_single_qubit_measurement_basis,
    parameterized_single_qubit_state_preparation,
    parameterized_readout_symmetrization,
)
from pyquil.gates import MEASURE, RX, RZ


def test_measure_qubits():
    p = Program()
    ro = p.declare("ro", "BIT", 2)
    p += MEASURE(0, ro[0])
    p += MEASURE(1, ro[1])
    assert measure_qubits([0, 1]).out() == p.out()


def test_parameterized_single_qubit_measurement_basis():
    p = Program()
    alpha = p.declare("measurement_alpha", "REAL", 2)
    beta = p.declare("measurement_beta", "REAL", 2)
    gamma = p.declare("measurement_gamma", "REAL", 2)
    for idx, q in enumerate(range(2)):
        p += RZ(alpha[idx], q)
        p += RX(np.pi / 2, q)
        p += RZ(beta[idx], q)
        p += RX(-np.pi / 2, q)
        p += RZ(gamma[idx], q)
    assert parameterized_single_qubit_measurement_basis([0, 1]).out() == p.out()


def test_parameterized_single_qubit_state_preparation():
    p = Program()
    alpha = p.declare("preparation_alpha", "REAL", 2)
    beta = p.declare("preparation_beta", "REAL", 2)
    gamma = p.declare("preparation_gamma", "REAL", 2)
    p += RZ(alpha[0], 0)
    p += RX(np.pi / 2, 0)
    p += RZ(beta[0], 0)
    p += RX(-np.pi / 2, 0)
    p += RZ(gamma[0], 0)
    p += RZ(alpha[1], 1)
    p += RX(np.pi / 2, 1)
    p += RZ(beta[1], 1)
    p += RX(-np.pi / 2, 1)
    p += RZ(gamma[1], 1)
    assert parameterized_single_qubit_state_preparation([0, 1]).out() == p.out()


def test_parameterized_readout_symmetrization():
    p = Program()
    symmetrization = p.declare("symmetrization", "REAL", 2)
    p += RX(symmetrization[0], 0)
    p += RX(symmetrization[1], 1)
    assert parameterized_readout_symmetrization([0, 1]).out() == p.out()



import networkx as nx
from pyquil.device.transformers.graph_to_compiler_isa import (
    compiler_isa_to_graph,
    DEFAULT_2Q_GATES,
    DEFAULT_1Q_GATES,
)
from pyquil.device.transformers import graph_to_compiler_isa
from pyquil.device.graph import NxDevice
from pyquil.external.rpcq import CompilerISA
from typing import Dict, Any


def test_isa_from_graph_order():
    """
    Test that edge keys in the CompilerISA are properly formatted and sorted by
    the node id.
    """
    # since node 16 appears first, even though we ask for the edge (15,16) the networkx internal
    # representation will have it as (16,15)
    fc = nx.from_edgelist([(16, 17), (15, 16)])
    isa = graph_to_compiler_isa(fc)
    isad = isa.dict(by_alias=True)
    for k in isad["2Q"]:
        q1, q2 = k.split("-")
        assert q1 < q2


def test_compiler_isa_to_graph(compiler_isa: CompilerISA):
    """
    Test that compiler_isa_to_graph transforms a ``CompilerISA`` to an ``nx.Graph``
    accurately and that an ``NxDevice.qubit_topology`` is isomorphic to the
    raw ``nx.Graph``.
    """
    graph = compiler_isa_to_graph(compiler_isa)
    should_be = nx.from_edgelist([(0, 1), (1, 2), (0, 2), (0, 3)])
    assert nx.is_isomorphic(graph, should_be)

    nx_device = NxDevice(graph)
    assert nx.is_isomorphic(graph, nx_device.qubit_topology())


def test_graph_to_compiler_isa(compiler_isa: CompilerISA, noise_model_dict: Dict[str, Any]):
    """
    Test that ```compiler_isa_to_graph``` transforms ``CompilerISA`` to an ``nx.Graph`` and that
    ``graph_to_compiler_isa`` transforms an ``nx.Device`` to a ``CompilerISA`` with default
    gates.
    """
    graph = compiler_isa_to_graph(compiler_isa)
    isa = graph_to_compiler_isa(graph)

    for _, qubit in isa.qubits.items():
        for gate in DEFAULT_1Q_GATES:
            assert any([gate == qubit_gate.operator for qubit_gate in qubit.gates])

    for _, edge in isa.edges.items():
        for gate in DEFAULT_2Q_GATES:
            assert any([gate == edge_gate.operator for edge_gate in edge.gates])



#!/usr/bin/python
##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import re
from math import pi

import numpy as np
import pytest

from pyquil.gates import (
    I,
    X,
    Y,
    Z,
    H,
    T,
    S,
    RX,
    RY,
    RZ,
    CNOT,
    CCNOT,
    PHASE,
    CPHASE00,
    CPHASE01,
    CPHASE10,
    CPHASE,
    SWAP,
    CSWAP,
    ISWAP,
    PSWAP,
    MEASURE,
    HALT,
    WAIT,
    NOP,
    RESET,
    NOT,
    AND,
    OR,
    MOVE,
    EXCHANGE,
    LOAD,
    CONVERT,
    STORE,
    XOR,
    IOR,
    NEG,
    ADD,
    SUB,
    MUL,
    DIV,
    EQ,
    GT,
    GE,
    LT,
    LE,
)
from pyquil.paulis import exponential_map, sZ
from pyquil.quil import (
    Program,
    merge_programs,
    merge_with_pauli_noise,
    address_qubits,
    get_classical_addresses_from_program,
    Pragma,
    validate_protoquil,
    validate_supported_quil,
)
from pyquil.quilatom import MemoryReference, Parameter, QubitPlaceholder, Sub, quil_cos, quil_sin
from pyquil.quilbase import (
    DefGate,
    Gate,
    Qubit,
    JumpWhen,
    Declare,
    DefCalibration,
    ClassicalNot,
    DefPermutationGate,
)
from pyquil.tests.utils import parse_equals


def test_gate():
    tg = Gate("TEST", qubits=[Qubit(1), Qubit(2)], params=[])
    assert tg.out() == "TEST 1 2"


def test_defgate():
    dg = DefGate("TEST", np.array([[1.0, 0.0], [0.0, 1.0]]))
    assert dg.out() == "DEFGATE TEST:\n    1.0, 0\n    0, 1.0\n"
    test = dg.get_constructor()
    tg = test(Qubit(1), Qubit(2))
    assert tg.out() == "TEST 1 2"
    tg = test(1, 2)
    assert tg.out() == "TEST 1 2"


def test_defgate_non_square_should_throw_error():
    with pytest.raises(ValueError) as error_info:
        DefGate("TEST", np.array([[0 + 0.5j, 0.5, 1], [0.5, 0 - 0.5j, 1]]))
    assert str(error_info.value) == "Matrix must be square."


def test_defgate_non_unitary_should_throw_error():
    with pytest.raises(ValueError) as error_info:
        DefGate("TEST", np.array([[0, 1], [2, 3]]))
    assert str(error_info.value) == "Matrix must be unitary."


def test_defgate_param():
    dgp = DefGate("TEST", [[1.0, 0.0], [0.0, 1.0]])
    assert dgp.out() == "DEFGATE TEST:\n    1.0, 0\n    0, 1.0\n"
    test = dgp.get_constructor()
    tg = test(Qubit(1))
    assert tg.out() == "TEST 1"


def test_inst_gates():
    p = Program()
    p.inst(H(0), X(1))
    assert len(p) == 2
    assert p.out() == "H 0\nX 1\n"


def test_inst_tuple():
    p = Program()
    p.inst(("Y", 0), ("X", 1))
    assert len(p) == 2
    assert p.out() == "Y 0\nX 1\n"


def test_inst_string():
    p = Program()
    p.inst("Y 0", "X 1")
    assert len(p) == 2
    assert p.out() == "Y 0\nX 1\n"


def test_program_pop():
    prog = Program(X(0), X(1))
    instruction = prog.pop()
    assert prog.out() == "X 0\n"
    assert Program(instruction).out() == "X 1\n"


def test_len_zero():
    prog = Program()
    assert len(prog) == 0


def test_len_one():
    prog = Program(X(0))
    assert len(prog) == 1


def test_len_nested():
    p = Program(Declare("ro", "BIT"), H(0)).measure(0, MemoryReference("ro", 0))
    q = Program(H(0), CNOT(0, 1))
    p.if_then(MemoryReference("ro", 0), q)
    assert len(p) == 9


def test_plus_operator():
    p = Program()
    p += H(0)
    p += [X(0), Y(0), Z(0)]
    assert len(p) == 4
    assert p.out() == "H 0\nX 0\nY 0\nZ 0\n"


def test_indexing():
    program = (
        Program(Declare("ro", "BIT"), H(0), Y(1), CNOT(0, 1))
        .measure(0, MemoryReference("ro", 0))
        .if_then(MemoryReference("ro", 0), Program(X(0)), Program())
    )
    assert program[1] == H(0)
    for ii, instr in enumerate(program.instructions):
        assert program[ii] == instr


def test_iteration():
    gate_list = [H(0), Y(1), CNOT(0, 1)]
    program = Program(gate_list)
    for ii, instruction in enumerate(program):
        assert instruction == gate_list[ii]

    # https://github.com/rigetti/pyquil/issues/265
    gate_generator = (gate_list[ii] for ii in range(3))
    program = Program(gate_generator)
    for ii, instruction in enumerate(program):
        assert instruction == gate_list[ii]


def test_program_plus_program():
    p = Program().inst(X(0))
    q = Program().inst(Y(0))
    r = p + q
    assert len(p.instructions) == 1
    assert len(q.instructions) == 1
    assert len(r.instructions) == 2
    assert p.out() == "X 0\n"
    assert q.out() == "Y 0\n"
    assert r.out() == "X 0\nY 0\n"


def test_program_tuple():
    p = Program()
    p.inst(("Y", 0), ("X", 1))
    assert len(p) == 2
    assert p.out() == "Y 0\nX 1\n"


def test_program_string():
    p = Program()
    p.inst("Y 0", "X 1")
    assert len(p.instructions) == 2
    assert p.instructions == [Y(0), X(1)]
    assert p.out() == "Y 0\nX 1\n"


def test_program_slice():
    p = Program(H(0), CNOT(0, 1), H(1))
    assert isinstance(p[1:], Program)


def test_prog_init():
    p = Program()
    p.inst(Declare("ro", "BIT"), X(0)).measure(0, MemoryReference("ro", 0))
    assert p.out() == ("DECLARE ro BIT[1]\nX 0\nMEASURE 0 ro[0]\n")


def test_classical_regs_implicit_ro():
    p = Program()
    p.inst(Declare("reg", "BIT", 2), X(0)).measure(0, MemoryReference("reg", 1))
    assert p.out() == "DECLARE reg BIT[2]\nX 0\nMEASURE 0 reg[1]\n"
    assert p.declarations == {
        "ro": Declare("ro", "BIT", 1),
        "reg": Declare("reg", "BIT", 2),
    }


def test_classical_regs_explicit_ro():
    p = Program()
    p.inst(Declare("ro", "BIT", 2), Declare("reg", "BIT", 2), X(0)).measure(0, MemoryReference("reg", 1))
    assert p.out() == "DECLARE ro BIT[2]\nDECLARE reg BIT[2]\nX 0\nMEASURE 0 reg[1]\n"
    assert p.declarations == {
        "ro": Declare("ro", "BIT", 2),
        "reg": Declare("reg", "BIT", 2),
    }


def test_simple_instructions():
    p = Program().inst(HALT, WAIT, RESET(), NOP)
    assert p.out() == "HALT\nWAIT\nRESET\nNOP\n"


def test_unary_classicals():
    p = Program()
    p.inst(
        MOVE(MemoryReference("ro", 0), 1),
        MOVE(MemoryReference("ro", 1), 0),
        NOT(MemoryReference("ro", 2)),
        NEG(MemoryReference("ro", 3)),
    )
    assert p.out() == "MOVE ro[0] 1\nMOVE ro[1] 0\nNOT ro[2]\nNEG ro[3]\n"


def test_binary_classicals():
    p = Program()

    # OR is deprecated in favor of IOR
    with pytest.warns(UserWarning):
        p.inst(OR(MemoryReference("ro", 1), MemoryReference("ro", 0)))

    p.inst(
        AND(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        MOVE(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        CONVERT(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        IOR(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        XOR(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        ADD(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        SUB(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        MUL(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        DIV(MemoryReference("ro", 0), MemoryReference("ro", 1)),
        EXCHANGE(MemoryReference("ro", 0), MemoryReference("ro", 1)),
    )

    assert (
        p.out() == "IOR ro[0] ro[1]\n"
        "AND ro[0] ro[1]\n"
        "MOVE ro[0] ro[1]\n"
        "CONVERT ro[0] ro[1]\n"
        "IOR ro[0] ro[1]\n"
        "XOR ro[0] ro[1]\n"
        "ADD ro[0] ro[1]\n"
        "SUB ro[0] ro[1]\n"
        "MUL ro[0] ro[1]\n"
        "DIV ro[0] ro[1]\n"
        "EXCHANGE ro[0] ro[1]\n"
    )


def test_memory_reference_unpacking():
    p = Program()

    p.inst(
        AND("ro", ("ro", 1)),
        MOVE("ro", ("ro", 1)),
        CONVERT("ro", ("ro", 1)),
        IOR("ro", ("ro", 1)),
        XOR("ro", ("ro", 1)),
        ADD("ro", ("ro", 1)),
        SUB("ro", ("ro", 1)),
        MUL("ro", ("ro", 1)),
        DIV("ro", ("ro", 1)),
        EXCHANGE("ro", ("ro", 1)),
    )

    assert (
        p.out() == "AND ro[0] ro[1]\n"
        "MOVE ro[0] ro[1]\n"
        "CONVERT ro[0] ro[1]\n"
        "IOR ro[0] ro[1]\n"
        "XOR ro[0] ro[1]\n"
        "ADD ro[0] ro[1]\n"
        "SUB ro[0] ro[1]\n"
        "MUL ro[0] ro[1]\n"
        "DIV ro[0] ro[1]\n"
        "EXCHANGE ro[0] ro[1]\n"
    )


def test_ternary_classicals():
    p = Program()
    p.inst(
        LOAD(MemoryReference("ro", 0), "ro", MemoryReference("n", 0)),
        STORE("ro", MemoryReference("n", 0), MemoryReference("ro", 0)),
        STORE("ro", MemoryReference("n", 0), 0),
        STORE("ro", MemoryReference("n", 0), 0.1),
        EQ(MemoryReference("ro", 0), MemoryReference("ro", 1), 0),
        EQ(MemoryReference("ro", 0), MemoryReference("ro", 1), 0.0),
        EQ(MemoryReference("ro", 0), MemoryReference("ro", 1), MemoryReference("ro", 0)),
        GE(MemoryReference("ro", 0), MemoryReference("ro", 1), 1),
        GE(MemoryReference("ro", 0), MemoryReference("ro", 1), 1.1),
        GE(MemoryReference("ro", 0), MemoryReference("ro", 1), MemoryReference("ro", 1)),
        GT(MemoryReference("ro", 0), MemoryReference("ro", 1), 2),
        GT(MemoryReference("ro", 0), MemoryReference("ro", 1), 2.2),
        GT(MemoryReference("ro", 0), MemoryReference("ro", 1), MemoryReference("ro", 2)),
        LE(MemoryReference("ro", 0), MemoryReference("ro", 1), 3),
        LE(MemoryReference("ro", 0), MemoryReference("ro", 1), 3.3),
        LE(MemoryReference("ro", 0), MemoryReference("ro", 1), MemoryReference("ro", 3)),
        LT(MemoryReference("ro", 0), MemoryReference("ro", 1), 4),
        LT(MemoryReference("ro", 0), MemoryReference("ro", 1), 4.4),
        LT(MemoryReference("ro", 0), MemoryReference("ro", 1), MemoryReference("ro", 4)),
    )
    assert (
        p.out() == "LOAD ro[0] ro n[0]\n"
        "STORE ro n[0] ro[0]\n"
        "STORE ro n[0] 0\n"
        "STORE ro n[0] 0.1\n"
        "EQ ro[0] ro[1] 0\n"
        "EQ ro[0] ro[1] 0.0\n"
        "EQ ro[0] ro[1] ro[0]\n"
        "GE ro[0] ro[1] 1\n"
        "GE ro[0] ro[1] 1.1\n"
        "GE ro[0] ro[1] ro[1]\n"
        "GT ro[0] ro[1] 2\n"
        "GT ro[0] ro[1] 2.2\n"
        "GT ro[0] ro[1] ro[2]\n"
        "LE ro[0] ro[1] 3\n"
        "LE ro[0] ro[1] 3.3\n"
        "LE ro[0] ro[1] ro[3]\n"
        "LT ro[0] ro[1] 4\n"
        "LT ro[0] ro[1] 4.4\n"
        "LT ro[0] ro[1] ro[4]\n"
    )


def test_measurement_calls():
    p = Program()
    p.inst(
        Declare("ro", "BIT", 2),
        MEASURE(0, MemoryReference("ro", 1)),
        MEASURE(0, MemoryReference("ro", 1)),
    )
    assert p.out() == ("DECLARE ro BIT[2]\nMEASURE 0 ro[1]\nMEASURE 0 ro[1]\n")


def test_measure_all():
    p = Program()
    mem = p.declare("ro", memory_size=4)
    p.measure_all((0, mem[0]), (1, mem[1]), (2, mem[3]))
    assert p.out() == "DECLARE ro BIT[4]\nMEASURE 0 ro[0]\nMEASURE 1 ro[1]\nMEASURE 2 ro[3]\n"

    p = Program([H(idx) for idx in range(4)])
    p.measure_all()
    for idx in range(4):
        assert p[idx + 5] == MEASURE(idx, MemoryReference("ro", idx))

    p = Program()
    p.measure_all()
    assert p.out() == ""


def test_dagger():
    p = Program(X(0), H(0))
    assert p.dagger().out() == "DAGGER H 0\nDAGGER X 0\n"

    p = Program(X(0), MEASURE(0, MemoryReference("ro", 0)))
    with pytest.raises(ValueError):
        p.dagger().out()

    # ensure that modifiers are preserved https://github.com/rigetti/pyquil/pull/914
    p = Program()
    control = 0
    target = 1
    cnot_control = 2
    p += X(target).controlled(control)
    p += Y(target).controlled(control)
    p += Z(target).controlled(control)
    p += H(target).controlled(control)
    p += S(target).controlled(control)
    p += T(target).controlled(control)
    p += PHASE(pi, target).controlled(control)
    p += CNOT(cnot_control, target).controlled(control)

    for instr, instr_dagger in zip(reversed(p._instructions), p.dagger()._instructions):
        assert "DAGGER " + instr.out() == instr_dagger.out()


def test_construction_syntax():
    p = Program().inst(Declare("ro", "BIT", 2), X(0), Y(1), Z(0)).measure(0, MemoryReference("ro", 1))
    assert p.out() == ("DECLARE ro BIT[2]\nX 0\nY 1\nZ 0\nMEASURE 0 ro[1]\n")
    p = (
        Program()
        .inst(Declare("ro", "BIT", 3), X(0))
        .inst(Y(1))
        .measure(0, MemoryReference("ro", 1))
        .inst(MEASURE(1, MemoryReference("ro", 2)))
    )
    assert p.out() == ("DECLARE ro BIT[3]\nX 0\nY 1\nMEASURE 0 ro[1]\nMEASURE 1 ro[2]\n")
    p = (
        Program()
        .inst(Declare("ro", "BIT", 2), X(0))
        .measure(0, MemoryReference("ro", 1))
        .inst(Y(1), X(0))
        .measure(0, MemoryReference("ro", 0))
    )
    assert p.out() == ("DECLARE ro BIT[2]\nX 0\nMEASURE 0 ro[1]\nY 1\nX 0\nMEASURE 0 ro[0]\n")


def test_singles():
    p = Program(I(0), X(0), Y(1), Z(1), H(2), T(2), S(1))
    assert p.out() == "I 0\nX 0\nY 1\nZ 1\nH 2\nT 2\nS 1\n"


def test_rotations():
    p = Program(RX(0.5, 0), RY(0.1, 1), RZ(1.4, 2))
    assert p.out() == "RX(0.5) 0\nRY(0.1) 1\nRZ(1.4) 2\n"


def test_controlled_gates():
    p = Program(CNOT(0, 1), CCNOT(0, 1, 2))
    assert p.out() == "CNOT 0 1\nCCNOT 0 1 2\n"


def test_phases():
    p = Program(
        PHASE(np.pi, 1),
        CPHASE00(np.pi, 0, 1),
        CPHASE01(np.pi, 0, 1),
        CPHASE10(np.pi, 0, 1),
        CPHASE(np.pi, 0, 1),
    )
    assert p.out() == "PHASE(pi) 1\nCPHASE00(pi) 0 1\n" "CPHASE01(pi) 0 1\nCPHASE10(pi) 0 1\n" "CPHASE(pi) 0 1\n"


def test_swaps():
    p = Program(SWAP(0, 1), CSWAP(0, 1, 2), ISWAP(0, 1), PSWAP(np.pi, 0, 1))
    assert p.out() == "SWAP 0 1\nCSWAP 0 1 2\nISWAP 0 1\nPSWAP(pi) 0 1\n"


def test_def_gate():
    # First we define the new gate from a matrix
    sqrt_x = np.array([[0.5 + 0.5j, 0.5 - 0.5j], [0.5 - 0.5j, 0.5 + 0.5j]])
    p = Program().defgate("SQRT-X", sqrt_x)

    # Then we can use the new gate
    p.inst(("SQRT-X", 0))
    assert p.out() == "DEFGATE SQRT-X:\n    0.5+0.5i, 0.5-0.5i\n    0.5-0.5i, 0.5+0.5i\n\nSQRT-X 0\n"


def test_def_gate_with_parameters():
    theta = Parameter("theta")
    rx = np.array(
        [
            [quil_cos(theta / 2), -1j * quil_sin(theta / 2)],
            [-1j * quil_sin(theta / 2), quil_cos(theta / 2)],
        ]
    )

    p = Program().defgate("RX", rx, [theta])
    assert (
        p.out() == "DEFGATE RX(%theta):\n"
        "    COS(%theta/2), -i*SIN(%theta/2)\n"
        "    -i*SIN(%theta/2), COS(%theta/2)\n\n"
    )

    dg = DefGate("MY_RX", rx, [theta])
    MY_RX = dg.get_constructor()
    p = Program().inst(MY_RX(np.pi)(0))
    assert p.out() == "MY_RX(pi) 0\n"


def test_multiqubit_gate():
    # A multi-qubit defgate example
    x_gate_matrix = np.array(([0.0, 1.0], [1.0, 0.0]))
    sqrt_x = np.array([[0.5 + 0.5j, 0.5 - 0.5j], [0.5 - 0.5j, 0.5 + 0.5j]])
    x_sqrt_x = np.kron(sqrt_x, x_gate_matrix)
    p = Program().defgate("X-SQRT-X", x_sqrt_x)

    # Then we can use the new gate
    p.inst(("X-SQRT-X", 0, 1))

    assert (
        p.out() == "DEFGATE X-SQRT-X:\n    0.0, 0.5+0.5i, 0.0, 0.5-0.5i\n    "
        "0.5+0.5i, 0.0, 0.5-0.5i, 0.0\n    "
        "0.0, 0.5-0.5i, 0.0, 0.5+0.5i\n    "
        "0.5-0.5i, 0.0, 0.5+0.5i, 0.0\n\nX-SQRT-X 0 1\n"
    )


def test_define_qft():
    def qft3(q0, q1, q2):
        p = Program()
        p.inst(
            H(q2),
            CPHASE(pi / 2.0, q1, q2),
            H(1),
            CPHASE(pi / 4.0, q0, q2),
            CPHASE(pi / 2.0, q0, q1),
            H(q0),
            SWAP(q0, q2),
        )
        return p

    # I(2) is to force 3 qubits in state prep program.
    state_prep = Program().inst(X(0))

    prog = state_prep + qft3(0, 1, 2)
    output = prog.out()
    assert output == "X 0\nH 2\nCPHASE(pi/2) 1 2\nH 1\nCPHASE(pi/4) 0 " "2\nCPHASE(pi/2) 0 1\nH 0\nSWAP 0 2\n"


def test_control_flows():
    outer_loop = Program()
    classical_flag_register = outer_loop.declare("classical_flag_register", "BIT")
    outer_loop += MOVE(classical_flag_register, 1)  # initialize

    inner_loop = Program()
    inner_loop += Program(X(0), H(0))
    inner_loop += MEASURE(0, classical_flag_register)

    # run inner_loop in a loop until classical_flag_register is 0
    outer_loop.while_do(classical_flag_register, inner_loop)
    assert outer_loop.out() == "\n".join(
        [
            "DECLARE classical_flag_register BIT[1]",
            "MOVE classical_flag_register 1",
            "LABEL @START1",
            "JUMP-UNLESS @END2 classical_flag_register",
            "X 0",
            "H 0",
            "MEASURE 0 classical_flag_register",
            "JUMP @START1",
            "LABEL @END2",
            "",
        ]
    )


def test_control_flows_2():
    # create a program that branches based on the value of a classical register
    x_prog = Program(X(0))
    z_prog = Program()
    branch = (
        Program(Declare("ro", "BIT", 2), H(1))
        .measure(1, MemoryReference("ro", 1))
        .if_then(MemoryReference("ro", 1), x_prog, z_prog)
        .measure(0, MemoryReference("ro", 0))
    )
    assert branch.out() == (
        "DECLARE ro BIT[2]\n"
        "H 1\n"
        "MEASURE 1 ro[1]\n"
        "JUMP-WHEN @THEN1 ro[1]\n"
        "JUMP @END2\n"
        "LABEL @THEN1\n"
        "X 0\n"
        "LABEL @END2\n"
        "MEASURE 0 ro[0]\n"
    )


def test_if_option():
    p = (
        Program(Declare("ro", "BIT", 1), X(0))
        .measure(0, MemoryReference("ro", 0))
        .if_then(MemoryReference("ro", 0), Program(X(1)))
    )
    assert p.out() == (
        "DECLARE ro BIT[1]\n"
        "X 0\n"
        "MEASURE 0 ro[0]\n"
        "JUMP-WHEN @THEN1 ro[0]\n"
        "JUMP @END2\n"
        "LABEL @THEN1\n"
        "X 1\n"
        "LABEL @END2\n"
    )

    assert isinstance(p.instructions[3], JumpWhen)


def test_alloc_deprecated():
    p = Program()

    with pytest.warns(DeprecationWarning):
        p.alloc()


def test_qubit_placeholder():
    p = Program()

    p.inst(H(0))  # H 0

    q1 = QubitPlaceholder()  # q1 = 1
    q2 = QubitPlaceholder()  # q2 = 3

    p.inst(CNOT(q1, q2))  # CNOT 1 3

    p.inst(H(2))

    q3 = QubitPlaceholder()  # q3 = 4

    p.inst(X(q3))  # X 4

    with pytest.raises(RuntimeError) as e:
        _ = p.out()
    assert e.match(r"Qubit q\d+ has not been assigned an index")


def test_qubit_placeholder_2():
    p = Program()

    p.inst(H(0))  # H 0

    q1 = QubitPlaceholder()  # q1 = 1
    q2 = QubitPlaceholder()  # q2 = 3

    p.inst(CNOT(q1, q2))  # CNOT 1 3

    p.inst(H(2))

    q3 = QubitPlaceholder()  # q3 = 4

    p.inst(X(q3))  # X 4
    with pytest.raises(ValueError) as e:
        _ = address_qubits(p, {q1: 1, q2: 3, q3: 4})

    assert e.match("Your program mixes instantiated qubits with placeholders")


def test_qubit_placeholder_new():
    p = Program()

    q0 = QubitPlaceholder()
    p.inst(H(q0))  # H 0

    q1 = QubitPlaceholder()
    q2 = QubitPlaceholder()

    p.inst(CNOT(q1, q2))  # CNOT 1 3

    qxxx = QubitPlaceholder()
    p.inst(H(qxxx))

    q3 = QubitPlaceholder()

    p.inst(X(q3))  # X 4
    p = address_qubits(p, {q1: 1, q2: 3, q3: 4, q0: 0, qxxx: 2})

    assert p.out() == "H 0\nCNOT 1 3\nH 2\nX 4\n"


def test_multiaddress():
    p = Program()
    q0, q1 = [QubitPlaceholder() for _ in range(2)]
    p += exponential_map(sZ(q0) * sZ(q1))(0.5)

    map1 = {q0: 0, q1: 1}
    map2 = {q0: 9, q1: 10}

    p1 = address_qubits(p, map1)

    with pytest.raises(RuntimeError):
        _ = p.out()  # make sure the original isn't affected

    assert p1.out() == "CNOT 0 1\nRZ(1.0) 1\nCNOT 0 1\n"

    p2 = address_qubits(p, map2)
    assert p1.out() == "CNOT 0 1\nRZ(1.0) 1\nCNOT 0 1\n"
    assert p2.out() == "CNOT 9 10\nRZ(1.0) 10\nCNOT 9 10\n"


def test_multiple_instantiate():
    p = Program()
    q = QubitPlaceholder()
    p.inst(H(q))
    p = address_qubits(p)
    assert p.out() == "H 0\n"
    assert p.out() == "H 0\n"


def test_reuse_placeholder():
    p = Program()
    q1 = QubitPlaceholder()
    q2 = QubitPlaceholder()
    p.inst(H(q1))
    p.inst(H(q2))
    p.inst(CNOT(q1, q2))
    p = address_qubits(p)
    assert p.out() == "H 0\nH 1\nCNOT 0 1\n"


def test_pauli_gate():
    s = """DEFGATE U(%beta) p q AS PAULI-SUM:
    ZZ(-1*%beta/4) p q
    Z(%beta/4) p
    Z(%beta/4) q

DEFGATE V:
    1.0, 0
    0, 1.0

U(2.0) 1 0
"""
    p = Program(s)
    assert p.out() == s


def test_prog_merge():
    prog_0 = Program(X(0))
    prog_1 = Program(Y(0))
    assert merge_programs([prog_0, prog_1]).out() == (prog_0 + prog_1).out()
    test_def = DefGate("test", np.eye(2))
    TEST = test_def.get_constructor()
    prog_0.inst(test_def)
    prog_0.inst(TEST(0))
    prog_1.inst(test_def)
    prog_1.inst(TEST(0))
    assert (
        merge_programs([prog_0, prog_1]).out()
        == """DEFGATE test:
    1.0, 0
    0, 1.0

X 0
test 0
Y 0
test 0
"""
    )
    perm_def = DefPermutationGate("PERM", [0, 1, 3, 2])
    PERM = perm_def.get_constructor()
    prog_0.inst(perm_def)
    prog_0.inst(PERM(0, 1))
    prog_1.inst(perm_def)
    prog_1.inst(PERM(1, 0))
    assert (
        merge_programs([prog_0, prog_1]).out()
        == """DEFGATE PERM AS PERMUTATION:
    0, 1, 3, 2
DEFGATE test:
    1.0, 0
    0, 1.0

X 0
test 0
PERM 0 1
Y 0
test 0
PERM 1 0
"""
    )
    assert (
        merge_programs([Program("DECLARE ro BIT[1]"), Program("H 0"), Program("MEASURE 0 ro[0]")]).out()
        == """DECLARE ro BIT[1]
H 0
MEASURE 0 ro[0]
"""
    )

    q0 = QubitPlaceholder()
    q0_str = "{" + str(q0) + "}"
    p0 = Program(X(q0))
    p1 = Program(Z(q0))
    merged = merge_programs([p0, p1])
    assert (
        str(merged)
        == f"""X {q0_str}
Z {q0_str}
"""
    )
    assert (
        address_qubits(merged, {q0: 1}).out()
        == """X 1
Z 1
"""
    )
    q1 = QubitPlaceholder()
    p2 = Program(Z(q1))
    assert (
        address_qubits(merge_programs([p0, p2]), {q0: 1, q1: 2}).out()
        == """X 1
Z 2
"""
    )
    p0 = address_qubits(p0, {q0: 2})
    p1 = address_qubits(p1, {q0: 1})
    assert (
        merge_programs([p0, p1]).out()
        == """X 2
Z 1
"""
    )


def test_merge_with_pauli_noise():
    p = Program(X(0)).inst(Z(0))
    probs = [0.0, 1.0, 0.0, 0.0]
    merged = merge_with_pauli_noise(p, probs, [0])
    assert (
        merged.out()
        == """DEFGATE pauli_noise:
    1.0, 0
    0, 1.0

PRAGMA ADD-KRAUS pauli_noise 0 "(0.0 0.0 0.0 0.0)"
PRAGMA ADD-KRAUS pauli_noise 0 "(0.0 1.0 1.0 0.0)"
PRAGMA ADD-KRAUS pauli_noise 0 "(0.0 0.0 0.0 0.0)"
PRAGMA ADD-KRAUS pauli_noise 0 "(0.0 0.0 0.0 -0.0)"
X 0
pauli_noise 0
Z 0
pauli_noise 0
"""
    )


def test_get_qubits():
    pq = Program(Declare("ro", "BIT"), X(0), CNOT(0, 4), MEASURE(5, MemoryReference("ro", 0)))
    assert pq.get_qubits() == {0, 4, 5}

    q = [QubitPlaceholder() for _ in range(6)]
    pq = Program(Declare("ro", "BIT"), X(q[0]), CNOT(q[0], q[4]), MEASURE(q[5], MemoryReference("ro", 0)))
    qq = QubitPlaceholder()
    pq.inst(Y(q[2]), X(qq))
    assert address_qubits(pq).get_qubits() == {0, 1, 2, 3, 4}

    qubit_index = 1
    p = Program(("H", qubit_index))
    assert p.get_qubits() == {qubit_index}
    q1 = QubitPlaceholder()
    q2 = QubitPlaceholder()
    p.inst(("CNOT", q1, q2))
    with pytest.raises(ValueError) as e:
        _ = address_qubits(p).get_qubits()
    assert e.match("Your program mixes instantiated qubits with placeholders")


def test_get_qubit_placeholders():
    qs = QubitPlaceholder.register(8)
    pq = Program(Declare("ro", "BIT"), X(qs[0]), CNOT(qs[0], qs[4]), MEASURE(qs[5], MemoryReference("ro", 0)))
    assert pq.get_qubits() == {qs[i] for i in [0, 4, 5]}


def test_get_qubits_not_as_indices():
    pq = Program(Declare("ro", "BIT"), X(0), CNOT(0, 4), MEASURE(5, MemoryReference("ro", 0)))
    assert pq.get_qubits(indices=False) == {Qubit(i) for i in [0, 4, 5]}


def test_eq():
    p1 = Program()
    q1 = QubitPlaceholder()
    q2 = QubitPlaceholder()
    p1.inst([H(q1), CNOT(q1, q2)])
    p1 = address_qubits(p1)

    p2 = Program()
    p2.inst([H(0), CNOT(0, 1)])

    assert p1 == p2
    assert not p1 != p2


def test_kraus():
    pq = Program(X(0))
    pq.define_noisy_gate("X", (0,), [[[0.0, 1.0], [1.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]])
    pq.inst(X(1))
    pq.define_noisy_gate("X", (1,), [[[0.0, 1.0], [1.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]])

    ret = pq.out()
    assert (
        ret
        == """X 0
PRAGMA ADD-KRAUS X 0 "(0.0 1.0 1.0 0.0)"
PRAGMA ADD-KRAUS X 0 "(0.0 0.0 0.0 0.0)"
X 1
PRAGMA ADD-KRAUS X 1 "(0.0 1.0 1.0 0.0)"
PRAGMA ADD-KRAUS X 1 "(0.0 0.0 0.0 0.0)"
"""
    )
    # test error due to bad normalization
    with pytest.raises(ValueError):
        pq.define_noisy_gate("X", (0,), [[[0.0, 1.0], [1.0, 0.0]], [[0.0, 1.0], [1.0, 0.0]]])
    # test error due to bad shape of kraus op
    with pytest.raises(ValueError):
        pq.define_noisy_gate("X", (0,), [[[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]], [[0.0, 1.0], [1.0, 0.0]]])

    pq1 = Program(X(0))
    pq1.define_noisy_gate("X", (0,), [[[0.0, 1.0], [1.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]])
    pq2 = Program(X(1))
    pq2.define_noisy_gate("X", (1,), [[[0.0, 1.0], [1.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]])

    assert pq1 + pq2 == pq

    pq_nn = Program(X(0))
    pq_nn.no_noise()
    pq_nn.inst(X(1))

    assert (
        pq_nn.out()
        == """X 0
PRAGMA NO-NOISE
X 1
"""
    )


def test_define_noisy_readout():
    pq = Program(X(0))
    pq.define_noisy_readout(0, 0.8, 0.9)

    pq.inst(X(1))
    pq.define_noisy_readout(1, 0.9, 0.8)

    ret = pq.out()
    assert (
        ret
        == """X 0
PRAGMA READOUT-POVM 0 "(0.8 0.09999999999999998 0.19999999999999996 0.9)"
X 1
PRAGMA READOUT-POVM 1 "(0.9 0.19999999999999996 0.09999999999999998 0.8)"
"""
    )
    # test error due to bad normalization
    with pytest.raises(ValueError):
        pq.define_noisy_readout(0, 1.1, 0.5)
    # test error due to bad normalization
    with pytest.raises(ValueError):
        pq.define_noisy_readout(0, 0.5, 1.5)
    # test error due to negative probability
    with pytest.raises(ValueError):
        pq.define_noisy_readout(0, -0.1, 0.5)
    # test error due to negative probability
    with pytest.raises(ValueError):
        pq.define_noisy_readout(0, 0.5, -1.0)
    # test error due to bad qubit_index value
    with pytest.raises(ValueError):
        pq.define_noisy_readout(-1, 0.5, 0.5)
    # test error due to bad qubit_index type
    with pytest.raises(TypeError):
        pq.define_noisy_readout(1.0, 0.5, 0.5)


# https://github.com/rigetti/pyquil/issues/72
def test_if_then_inherits_defined_gates():
    p1 = Program()
    p1.inst(H(0))
    p1.measure(0, MemoryReference("ro", 0))

    p2 = Program()
    p2.defgate("A", np.array([[1.0, 0.0], [0.0, 1.0]]))
    p2.inst(("A", 0))

    p3 = Program()
    p3.defgate("B", np.array([[0.0, 1.0], [1.0, 0.0]]))
    p3.inst(("B", 0))

    p1.if_then(MemoryReference("ro", 0), p2, p3)
    assert p2.defined_gates[0] in p1.defined_gates
    assert p3.defined_gates[0] in p1.defined_gates


# https://github.com/rigetti/pyquil/issues/124
def test_allocating_qubits_on_multiple_programs():
    p = Program()
    qubit0 = QubitPlaceholder()
    p.inst(X(qubit0))

    q = Program()
    qubit1 = QubitPlaceholder()
    q.inst(X(qubit1))

    assert address_qubits(p + q).out() == "X 0\nX 1\n"


# https://github.com/rigetti/pyquil/issues/163
def test_installing_programs_inside_other_programs():
    p = Program()
    q = Program()
    p.inst(q)
    assert len(p) == 0


# https://github.com/rigetti/pyquil/issues/168
def test_nesting_a_program_inside_itself():
    p = Program(H(0)).measure(0, MemoryReference("ro", 0))
    with pytest.raises(ValueError):
        p.if_then(MemoryReference("ro", 0), p)


# https://github.com/rigetti/pyquil/issues/170
def test_inline_placeholder():
    p = Program()
    p += H(QubitPlaceholder())
    assert address_qubits(p).out() == "H 0\n"


# https://github.com/rigetti/pyquil/issues/138
def test_defgate_integer_input():
    dg = DefGate("TEST", np.array([[1, 0], [0, 1]]))
    assert dg.out() == "DEFGATE TEST:\n    1, 0\n    0, 1\n"


def test_out_vs_str():
    qs = QubitPlaceholder.register(6)
    pq = Program(
        Declare("ro", "BIT", 6),
        X(qs[0]),
        CNOT(qs[0], qs[4]),
        MEASURE(qs[5], MemoryReference("ro", 5)),
    )

    with pytest.raises(RuntimeError) as e:
        pq.out()
    assert e.match(r"Qubit q\d+ has not been assigned an index")

    string_version = str(pq)
    should_be_re = r"DECLARE ro BIT\[6\]\nX \{q\d+\}\nCNOT \{q\d+\} \{q\d+\}\nMEASURE \{q\d+\} ro\[5\]\n"
    assert re.fullmatch(should_be_re, string_version, flags=re.MULTILINE)


def test_get_classical_addresses_from_program():
    p = Program(Declare("ro", "BIT", 4), [H(i) for i in range(4)])
    assert get_classical_addresses_from_program(p) == {}

    p += [MEASURE(i, MemoryReference("ro", i)) for i in [0, 3, 1]]
    assert get_classical_addresses_from_program(p) == {"ro": [0, 1, 3]}


def test_get_classical_addresses_from_quil_program():
    """
    Tests that can get_classical_addresses_from_program can handle both MEASURE
    quil instructions with and without explicit classical registers.
    """

    p = Program("\n".join(["X 0", "MEASURE 0"]))
    assert get_classical_addresses_from_program(p) == {}

    p = Program("\n".join(["DECLARE ro BIT[2]", "X 0", "MEASURE 0 ro[1]"]))
    assert get_classical_addresses_from_program(p) == {"ro": [1]}


def test_pragma_with_placeholders():
    q = QubitPlaceholder()
    q2 = QubitPlaceholder()
    p = Program()
    p.inst(Pragma("FENCE", [q, q2]))
    address_map = {q: 0, q2: 1}
    addressed_pragma = address_qubits(p, address_map)[0]
    parse_equals("PRAGMA FENCE 0 1\n", addressed_pragma)

    pq = Program(X(q))
    pq.define_noisy_readout(q, 0.8, 0.9)

    pq.inst(X(q2))
    pq.define_noisy_readout(q2, 0.9, 0.8)

    ret = address_qubits(pq, address_map).out()
    assert (
        ret
        == """X 0
PRAGMA READOUT-POVM 0 "(0.8 0.09999999999999998 0.19999999999999996 0.9)"
X 1
PRAGMA READOUT-POVM 1 "(0.9 0.19999999999999996 0.09999999999999998 0.8)"
"""
    )


def test_implicit_declare():
    with pytest.warns(UserWarning):
        program = Program(MEASURE(0, MemoryReference("ro", 0)))
        assert program.out() == ("DECLARE ro BIT[1]\nMEASURE 0 ro[0]\n")


def test_no_implicit_declare():
    program = Program(Declare("read_out", "BIT", 5), MEASURE(0, MemoryReference("read_out", 4)))
    assert program.out() == ("DECLARE read_out BIT[5]\nMEASURE 0 read_out[4]\n")


def test_no_implicit_declare_2():
    program = Program(MEASURE(0, MemoryReference("asdf", 4)))
    assert program.out() == "MEASURE 0 asdf[4]\n"


def test_reset():
    p = Program()
    p.reset(0)
    p.reset()
    assert p.out() == "RESET 0\nRESET\n"

    program = Program()
    qubit = QubitPlaceholder()
    # address_qubits() won't work unless there's a gate besides
    # RESET on a QubitPlaceholder, this is just here to make
    # addressing work
    program += X(qubit)

    program += RESET(qubit)
    program = address_qubits(program)
    assert program.out() == "X 0\nRESET 0\n"


def test_copy():
    prog1 = Program(H(0), CNOT(0, 1))
    prog2 = prog1.copy().measure_all()
    assert prog1.out() == "\n".join(["H 0", "CNOT 0 1", ""])
    assert prog2.out() == "\n".join(["H 0", "CNOT 0 1", "DECLARE ro BIT[2]", "MEASURE 0 ro[0]", "MEASURE 1 ro[1]", ""])


def test_measure_all_noncontig():
    prog = Program(H(0), H(10))
    prog.measure_all()
    assert prog.out() == "\n".join(["H 0", "H 10", "DECLARE ro BIT[11]", "MEASURE 0 ro[0]", "MEASURE 10 ro[10]", ""])


def test_validate_supported_quil_reset_first():
    prog = Program(H(0), RESET())
    with pytest.raises(ValueError):
        validate_supported_quil(prog)
    assert not prog.is_supported_on_qpu()


def test_validate_supported_quil_reset_qubit():
    prog = Program(RESET(2))
    with pytest.raises(ValueError):
        validate_supported_quil(prog)
    assert not prog.is_supported_on_qpu()


def test_validate_supported_quil_measure_last():
    prog = Program(MEASURE(0, None), H(0))
    with pytest.raises(ValueError):
        validate_supported_quil(prog)
    assert not prog.is_supported_on_qpu()


def test_validate_supported_quil_with_pragma():
    prog = Program(RESET(), H(1), Pragma("DELAY"), MEASURE(1, None))
    assert prog.is_supported_on_qpu()


def test_validate_supported_quil_suite():
    validate_supported_quil(
        Program(
            """
RESET
DECLARE ro BIT[3]
RX(-pi/4) 2
RZ(4*pi) 3
I 0
CZ 2 3
MEASURE 2 ro[2]
MEASURE 3 ro[3]
"""
        )
    )

    validate_supported_quil(
        Program(
            """
RESET
DECLARE ro BIT[3]
RX(-pi/4) 2
RZ(4*pi) 3
I 0
CZ 2 3
MEASURE 2 ro[2]
MEASURE 3 ro[3]
"""
        )
    )
    validate_supported_quil(
        Program(
            """
RESET
DECLARE ro BIT[3]
RX(-pi/4) 2
RZ(4*pi) 3
I 0
MEASURE 0
CZ 2 3
MEASURE 2 ro[2]
X 3
MEASURE 3 ro[3]
"""
        )
    )

    with pytest.raises(ValueError):
        validate_supported_quil(
            Program(
                """
RESET
DECLARE ro BIT[3]
RX(-pi/4) 2
RZ(4*pi) 3
RESET
I 0
CZ 2 3
MEASURE 2 ro[2]
MEASURE 3 ro[3]
"""
            )
        )

    with pytest.raises(ValueError):
        validate_supported_quil(
            Program(
                """
RESET
DECLARE ro BIT[3]
RX(-pi/4) 2
RZ(4*pi) 3
MEASURE 2
I 0
CZ 2 3
MEASURE 2 ro[2]
MEASURE 3 ro[3]
"""
            )
        )

    with pytest.raises(ValueError):
        validate_supported_quil(
            Program(
                """
RESET
DECLARE ro BIT[3]
RX(-pi/4) 2
RZ(4*pi) 3
HALT
I 0
CZ 2 3
MEASURE 2 ro[2]
MEASURE 3 ro[3]
"""
            )
        )


def test_validate_supported_quil_multiple_measures():
    prog = Program(RESET(), H(1), Pragma("DELAY"), MEASURE(1, None), MEASURE(1, None))
    with pytest.raises(ValueError):
        validate_supported_quil(prog)


def test_is_protoquil():
    prog = Program(Declare("ro", "BIT"), MEASURE(1, MemoryReference("ro", 0)), H(1), RESET())
    validate_protoquil(prog)
    assert prog.is_protoquil()

    prog = (
        Program(Declare("ro", "BIT"), H(0), Y(1), CNOT(0, 1))
        .measure(0, MemoryReference("ro", 0))
        .if_then(MemoryReference("ro", 0), Program(X(0)), Program())
    )
    with pytest.raises(ValueError):
        validate_protoquil(prog)
    assert not prog.is_protoquil()

    prog = Program(Declare("ro", "BIT"), ClassicalNot(MemoryReference("ro", 0)))
    with pytest.raises(ValueError):
        validate_protoquil(prog)
    assert not prog.is_protoquil()

    prog = Program(DefCalibration("I", [], [Qubit(0)], []), I(0))
    with pytest.raises(ValueError):
        validate_protoquil(prog)
    assert not prog.is_protoquil()
    assert prog.is_protoquil(quilt=True)


def test_subtracting_memory_regions():
    # https://github.com/rigetti/pyquil/issues/709
    p = Program()
    alpha = p.declare("alpha", "REAL")
    beta = p.declare("beta", "REAL")
    p += RZ(alpha - beta, 0)
    p2 = Program(p.out())
    parsed_rz = p2.pop()  # type: Gate
    parsed_param = parsed_rz.params[0]
    assert isinstance(parsed_param, Sub)
    assert parsed_param.op1 == alpha
    assert parsed_param.op2 == beta


def test_out_of_bounds_memory():
    r = Program().declare("ro", "BIT", 1)
    with pytest.raises(IndexError):
        r[1]


@pytest.mark.timeout(5)
def test_memory_reference_iteration():
    r = Program().declare("ro", "BIT", 10)
    assert len([i for i in r]) == 10


def test_placeholders_preserves_modifiers():
    cs = QubitPlaceholder.register(3)
    ts = QubitPlaceholder.register(1)

    g = X(ts[0])
    for c in cs:
        g = g.controlled(c).dagger()

    p = Program(g)
    a = address_qubits(p)

    assert a[0].modifiers == g.modifiers


def _eval_as_np_pi(exp):
    return eval(exp.replace("pi", repr(np.pi)).replace("theta[0]", "1"))


def test_params_pi_and_precedence():
    trivial_pi = "3 * theta[0] / (2 * pi)"
    prog = Program(f"RX({trivial_pi}) 0")
    exp = str(prog[0].params[0])
    assert _eval_as_np_pi(trivial_pi) == _eval_as_np_pi(exp)

    less_trivial_pi = "3 * theta[0] * 2 / (pi)"
    prog = Program(f"RX({less_trivial_pi}) 0")
    exp = str(prog[0].params[0])
    assert _eval_as_np_pi(less_trivial_pi) == _eval_as_np_pi(exp)

    more_less_trivial_pi = "3 / (theta[0] / (pi + 1)) / pi"
    prog = Program(f"RX({more_less_trivial_pi}) 0")
    exp = str(prog[0].params[0])
    assert _eval_as_np_pi(more_less_trivial_pi) == _eval_as_np_pi(exp)



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for creating and verifying noisy gate and readout definitions.
"""
from collections import namedtuple
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, TYPE_CHECKING, Union

import numpy as np
import sys

from pyquil.external.rpcq import CompilerISA
from pyquil.gates import I, MEASURE, X
from pyquil.quilbase import Pragma, Gate
from pyquil.quilatom import MemoryReference, format_parameter, ParameterDesignator
from pyquil.noise_gates import _get_qvm_noise_supported_gates

if TYPE_CHECKING:
    from pyquil.quil import Program
    from pyquil.api import QPUConnection, QVMConnection  # noqa: F401

INFINITY = float("inf")
"Used for infinite coherence times."


_KrausModel = namedtuple("_KrausModel", ["gate", "params", "targets", "kraus_ops", "fidelity"])


class KrausModel(_KrausModel):
    """
    Encapsulate a single gate's noise model.

    :ivar str gate: The name of the gate.
    :ivar Sequence[float] params: Optional parameters for the gate.
    :ivar Sequence[int] targets: The target qubit ids.
    :ivar Sequence[np.array] kraus_ops: The Kraus operators (must be square complex numpy arrays).
    :ivar float fidelity: The average gate fidelity associated with the Kraus map relative to the
        ideal operation.
    """

    @staticmethod
    def unpack_kraus_matrix(m: Union[List[Any], np.ndarray]) -> np.ndarray:
        """
        Helper to optionally unpack a JSON compatible representation of a complex Kraus matrix.

        :param m: The representation of a Kraus operator. Either a complex
            square matrix (as numpy array or nested lists) or a JSON-able pair of real matrices
            (as nested lists) representing the element-wise real and imaginary part of m.
        :return: A complex square numpy array representing the Kraus operator.
        """
        m = np.asarray(m, dtype=complex)
        if m.ndim == 3:
            m = m[0] + 1j * m[1]
        if not m.ndim == 2:  # pragma no coverage
            raise ValueError("Need 2d array.")
        if not m.shape[0] == m.shape[1]:  # pragma no coverage
            raise ValueError("Need square matrix.")
        return m

    def to_dict(self) -> Dict[str, Any]:
        """
        Create a dictionary representation of a KrausModel.

        For example::

            {
                "gate": "RX",
                "params": np.pi,
                "targets": [0],
                "kraus_ops": [            # In this example single Kraus op = ideal RX(pi) gate
                    [[[0,   0],           # element-wise real part of matrix
                      [0,   0]],
                      [[0, -1],           # element-wise imaginary part of matrix
                      [-1, 0]]]
                ],
                "fidelity": 1.0
            }

        :return: A JSON compatible dictionary representation.
        :rtype: Dict[str,Any]
        """
        res = self._asdict()
        res["kraus_ops"] = [[k.real.tolist(), k.imag.tolist()] for k in self.kraus_ops]
        return res

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "KrausModel":
        """
        Recreate a KrausModel from the dictionary representation.

        :param d: The dictionary representing the KrausModel. See `to_dict` for an
            example.
        :return: The deserialized KrausModel.
        """
        kraus_ops = [KrausModel.unpack_kraus_matrix(k) for k in d["kraus_ops"]]
        return KrausModel(d["gate"], d["params"], d["targets"], kraus_ops, d["fidelity"])

    def __eq__(self, other: object) -> bool:
        return isinstance(other, KrausModel) and self.to_dict() == other.to_dict()

    def __neq__(self, other: object) -> bool:
        return not self.__eq__(other)


_NoiseModel = namedtuple("_NoiseModel", ["gates", "assignment_probs"])


class NoiseModel(_NoiseModel):
    """
    Encapsulate the QPU noise model containing information about the noisy gates.

    :ivar Sequence[KrausModel] gates: The tomographic estimates of all gates.
    :ivar Dict[int,np.array] assignment_probs: The single qubit readout assignment
        probability matrices keyed by qubit id.
    """

    def to_dict(self) -> Dict[str, Any]:
        """
        Create a JSON serializable representation of the noise model.

        For example::

            {
                "gates": [
                    # list of embedded dictionary representations of KrausModels here [...]
                ]
                "assignment_probs": {
                    "0": [[.8, .1],
                          [.2, .9]],
                    "1": [[.9, .4],
                          [.1, .6]],
                }
            }

        :return: A dictionary representation of self.
        """
        return {
            "gates": [km.to_dict() for km in self.gates],
            "assignment_probs": {str(qid): a.tolist() for qid, a in self.assignment_probs.items()},
        }

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NoiseModel":
        """
        Re-create the noise model from a dictionary representation.

        :param d: The dictionary representation.
        :return: The restored noise model.
        """
        return NoiseModel(
            gates=[KrausModel.from_dict(t) for t in d["gates"]],
            assignment_probs={int(qid): np.array(a) for qid, a in d["assignment_probs"].items()},
        )

    def gates_by_name(self, name: str) -> List[KrausModel]:
        """
        Return all defined noisy gates of a particular gate name.

        :param str name: The gate name.
        :return: A list of noise models representing that gate.
        """
        return [g for g in self.gates if g.gate == name]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NoiseModel) and self.to_dict() == other.to_dict()

    def __neq__(self, other: object) -> bool:
        return not self.__eq__(other)


def _check_kraus_ops(n: int, kraus_ops: Sequence[np.ndarray]) -> None:
    """
    Verify that the Kraus operators are of the correct shape and satisfy the correct normalization.

    :param n: Number of qubits
    :param kraus_ops: The Kraus operators as numpy.ndarrays.
    """
    for k in kraus_ops:
        if not np.shape(k) == (2 ** n, 2 ** n):
            raise ValueError("Kraus operators for {0} qubits must have shape {1}x{1}: {2}".format(n, 2 ** n, k))

    kdk_sum = sum(np.transpose(k).conjugate().dot(k) for k in kraus_ops)
    if not np.allclose(kdk_sum, np.eye(2 ** n), atol=1e-3):
        raise ValueError("Kraus operator not correctly normalized: sum_j K_j^*K_j == {}".format(kdk_sum))


def _create_kraus_pragmas(name: str, qubit_indices: Sequence[int], kraus_ops: Sequence[np.ndarray]) -> List[Pragma]:
    """
    Generate the pragmas to define a Kraus map for a specific gate on some qubits.

    :param name: The name of the gate.
    :param qubit_indices: The qubits
    :param kraus_ops: The Kraus operators as matrices.
    :return: A QUIL string with PRAGMA ADD-KRAUS ... statements.
    """

    pragmas = [
        Pragma(
            "ADD-KRAUS",
            (name,) + tuple(qubit_indices),
            "({})".format(" ".join(map(format_parameter, np.ravel(k)))),
        )
        for k in kraus_ops
    ]
    return pragmas


def append_kraus_to_gate(
    kraus_ops: Sequence[np.ndarray], gate_matrix: np.ndarray
) -> List[Union[np.number, np.ndarray]]:  # type: ignore
    """
    Follow a gate ``gate_matrix`` by a Kraus map described by ``kraus_ops``.

    :param kraus_ops: The Kraus operators.
    :param gate_matrix: The unitary gate.
    :return: A list of transformed Kraus operators.
    """
    return [kj.dot(gate_matrix) for kj in kraus_ops]


def pauli_kraus_map(probabilities: Sequence[float]) -> List[np.ndarray]:
    r"""
    Generate the Kraus operators corresponding to a pauli channel.

    :params probabilities: The 4^num_qubits list of probabilities specifying the
        desired pauli channel. There should be either 4 or 16 probabilities specified in the
        order I, X, Y, Z for 1 qubit or II, IX, IY, IZ, XI, XX, XY, etc for 2 qubits.

            For example::

                The d-dimensional depolarizing channel \Delta parameterized as
                \Delta(\rho) = p \rho + [(1-p)/d] I
                is specified by the list of probabilities
                [p + (1-p)/d, (1-p)/d,  (1-p)/d), ... , (1-p)/d)]

    :return: A list of the 4^num_qubits Kraus operators that parametrize the map.
    """
    if len(probabilities) not in [4, 16]:
        raise ValueError(
            "Currently we only support one or two qubits, "
            "so the provided list of probabilities must have length 4 or 16."
        )
    if not np.allclose(sum(probabilities), 1.0, atol=1e-3):
        raise ValueError("Probabilities must sum to one.")

    paulis = [
        np.eye(2),
        np.array([[0, 1], [1, 0]]),
        np.array([[0, -1j], [1j, 0]]),
        np.array([[1, 0], [0, -1]]),
    ]

    if len(probabilities) == 4:
        operators = paulis
    else:
        operators = np.kron(paulis, paulis)

    return [coeff * op for coeff, op in zip(np.sqrt(probabilities), operators)]


def damping_kraus_map(p: float = 0.10) -> List[np.ndarray]:
    """
    Generate the Kraus operators corresponding to an amplitude damping
    noise channel.

    :param p: The one-step damping probability.
    :return: A list [k1, k2] of the Kraus operators that parametrize the map.
    :rtype: list
    """
    damping_op = np.sqrt(p) * np.array([[0, 1], [0, 0]])

    residual_kraus = np.diag([1, np.sqrt(1 - p)])
    return [residual_kraus, damping_op]


def dephasing_kraus_map(p: float = 0.10) -> List[np.ndarray]:
    """
    Generate the Kraus operators corresponding to a dephasing channel.

    :params float p: The one-step dephasing probability.
    :return: A list [k1, k2] of the Kraus operators that parametrize the map.
    :rtype: list
    """
    return [np.sqrt(1 - p) * np.eye(2), np.sqrt(p) * np.diag([1, -1])]


def tensor_kraus_maps(k1: List[np.ndarray], k2: List[np.ndarray]) -> List[np.ndarray]:
    """
    Generate the Kraus map corresponding to the composition
    of two maps on different qubits.

    :param k1: The Kraus operators for the first qubit.
    :param k2: The Kraus operators for the second qubit.
    :return: A list of tensored Kraus operators.
    """
    return [np.kron(k1j, k2l) for k1j in k1 for k2l in k2]


def combine_kraus_maps(k1: List[np.ndarray], k2: List[np.ndarray]) -> List[np.ndarray]:
    """
    Generate the Kraus map corresponding to the composition
    of two maps on the same qubits with k1 being applied to the state
    after k2.

    :param k1: The list of Kraus operators that are applied second.
    :param k2: The list of Kraus operators that are applied first.
    :return: A combinatorially generated list of composed Kraus operators.
    """
    return [np.dot(k1j, k2l) for k1j in k1 for k2l in k2]


def damping_after_dephasing(T1: float, T2: float, gate_time: float) -> List[np.ndarray]:
    """
    Generate the Kraus map corresponding to the composition
    of a dephasing channel followed by an amplitude damping channel.

    :param T1: The amplitude damping time
    :param T2: The dephasing time
    :param gate_time: The gate duration.
    :return: A list of Kraus operators.
    """
    assert T1 >= 0
    assert T2 >= 0

    if T1 != INFINITY:
        damping = damping_kraus_map(p=1 - np.exp(-float(gate_time) / float(T1)))
    else:
        damping = [np.eye(2)]

    if T2 != INFINITY:
        gamma_phi = float(gate_time) / float(T2)
        if T1 != INFINITY:
            if T2 > 2 * T1:
                raise ValueError("T2 is upper bounded by 2 * T1")
            gamma_phi -= float(gate_time) / float(2 * T1)

        dephasing = dephasing_kraus_map(p=0.5 * (1 - np.exp(-gamma_phi)))
    else:
        dephasing = [np.eye(2)]
    return combine_kraus_maps(damping, dephasing)


# You can only apply gate-noise to non-parametrized gates or parametrized gates at fixed parameters.
NO_NOISE = ["RZ"]
ANGLE_TOLERANCE = 1e-10


class NoisyGateUndefined(Exception):
    """Raise when user attempts to use noisy gate outside of currently supported set."""

    pass


def get_noisy_gate(gate_name: str, params: Iterable[ParameterDesignator]) -> Tuple[np.ndarray, str]:
    """
    Look up the numerical gate representation and a proposed 'noisy' name.

    :param gate_name: The Quil gate name
    :param params: The gate parameters.
    :return: A tuple (matrix, noisy_name) with the representation of the ideal gate matrix
        and a proposed name for the noisy version.
    """
    params = tuple(params)
    if gate_name == "I":
        assert params == ()
        return np.eye(2), "NOISY-I"
    if gate_name == "RX":
        (angle,) = params
        if not isinstance(angle, (int, float, complex)):
            raise TypeError(f"Cannot produce noisy gate for parameter of type {type(angle)}")

        if np.isclose(angle, np.pi / 2, atol=ANGLE_TOLERANCE):
            return np.array([[1, -1j], [-1j, 1]]) / np.sqrt(2), "NOISY-RX-PLUS-90"
        elif np.isclose(angle, -np.pi / 2, atol=ANGLE_TOLERANCE):
            return np.array([[1, 1j], [1j, 1]]) / np.sqrt(2), "NOISY-RX-MINUS-90"
        elif np.isclose(angle, np.pi, atol=ANGLE_TOLERANCE):
            return np.array([[0, -1j], [-1j, 0]]), "NOISY-RX-PLUS-180"
        elif np.isclose(angle, -np.pi, atol=ANGLE_TOLERANCE):
            return np.array([[0, 1j], [1j, 0]]), "NOISY-RX-MINUS-180"
    elif gate_name == "CZ":
        assert params == ()
        return np.diag([1, 1, 1, -1]), "NOISY-CZ"

    raise NoisyGateUndefined(
        "Undefined gate and params: {}{}\n"
        "Please restrict yourself to I, RX(+/-pi), RX(+/-pi/2), CZ".format(gate_name, params)
    )


def _get_program_gates(prog: "Program") -> List[Gate]:
    """
    Get all gate applications appearing in prog.

    :param prog: The program
    :return: A list of all Gates in prog (without duplicates).
    """
    return sorted({i for i in prog if isinstance(i, Gate)}, key=lambda g: g.out())


def _decoherence_noise_model(
    gates: Sequence[Gate],
    T1: Union[Dict[int, float], float] = 30e-6,
    T2: Union[Dict[int, float], float] = 30e-6,
    gate_time_1q: float = 50e-9,
    gate_time_2q: float = 150e-09,
    ro_fidelity: Union[Dict[int, float], float] = 0.95,
) -> NoiseModel:
    """
    The default noise parameters

    - T1 = 30 us
    - T2 = 30 us
    - 1q gate time = 50 ns
    - 2q gate time = 150 ns

    are currently typical for near-term devices.

    This function will define new gates and add Kraus noise to these gates. It will translate
    the input program to use the noisy version of the gates.

    :param gates: The gates to provide the noise model for.
    :param T1: The T1 amplitude damping time either globally or in a
        dictionary indexed by qubit id. By default, this is 30 us.
    :param T2: The T2 dephasing time either globally or in a
        dictionary indexed by qubit id. By default, this is also 30 us.
    :param gate_time_1q: The duration of the one-qubit gates, namely RX(+pi/2) and RX(-pi/2).
        By default, this is 50 ns.
    :param gate_time_2q: The duration of the two-qubit gates, namely CZ.
        By default, this is 150 ns.
    :param ro_fidelity: The readout assignment fidelity
        :math:`F = (p(0|0) + p(1|1))/2` either globally or in a dictionary indexed by qubit id.
    :return: A NoiseModel with the appropriate Kraus operators defined.
    """
    all_qubits = set(sum(([t.index for t in g.qubits] for g in gates), []))
    if isinstance(T1, dict):
        all_qubits.update(T1.keys())
    if isinstance(T2, dict):
        all_qubits.update(T2.keys())
    if isinstance(ro_fidelity, dict):
        all_qubits.update(ro_fidelity.keys())

    if not isinstance(T1, dict):
        T1 = {q: T1 for q in all_qubits}

    if not isinstance(T2, dict):
        T2 = {q: T2 for q in all_qubits}

    if not isinstance(ro_fidelity, dict):
        ro_fidelity = {q: ro_fidelity for q in all_qubits}

    noisy_identities_1q = {
        q: damping_after_dephasing(T1.get(q, INFINITY), T2.get(q, INFINITY), gate_time_1q) for q in all_qubits
    }
    noisy_identities_2q = {
        q: damping_after_dephasing(T1.get(q, INFINITY), T2.get(q, INFINITY), gate_time_2q) for q in all_qubits
    }
    kraus_maps = []
    for g in gates:
        targets = tuple(t.index for t in g.qubits)
        if g.name in NO_NOISE:
            continue
        matrix, _ = get_noisy_gate(g.name, g.params)

        if len(targets) == 1:
            noisy_I = noisy_identities_1q[targets[0]]
        else:
            if len(targets) != 2:
                raise ValueError("Noisy gates on more than 2Q not currently supported")

            # note this ordering of the tensor factors is necessary due to how the QVM orders
            # the wavefunction basis
            noisy_I = tensor_kraus_maps(noisy_identities_2q[targets[1]], noisy_identities_2q[targets[0]])
        kraus_maps.append(
            KrausModel(
                g.name,
                tuple(g.params),
                targets,
                combine_kraus_maps(noisy_I, [matrix]),
                # FIXME (Nik): compute actual avg gate fidelity for this simple
                # noise model
                1.0,
            )
        )
    aprobs = {}
    for q, f_ro in ro_fidelity.items():
        aprobs[q] = np.array([[f_ro, 1.0 - f_ro], [1.0 - f_ro, f_ro]])

    return NoiseModel(kraus_maps, aprobs)


def decoherence_noise_with_asymmetric_ro(isa: CompilerISA, p00: float = 0.975, p11: float = 0.911) -> NoiseModel:
    """Similar to :py:func:`_decoherence_noise_model`, but with asymmetric readout.

    For simplicity, we use the default values for T1, T2, gate times, et al. and only allow
    the specification of readout fidelities.
    """
    gates = _get_qvm_noise_supported_gates(isa)
    noise_model = _decoherence_noise_model(gates)
    aprobs = np.array([[p00, 1 - p00], [1 - p11, p11]])
    aprobs = {q: aprobs for q in noise_model.assignment_probs.keys()}
    return NoiseModel(noise_model.gates, aprobs)


def _noise_model_program_header(noise_model: NoiseModel) -> "Program":
    """
    Generate the header for a pyquil Program that uses ``noise_model`` to overload noisy gates.
    The program header consists of 3 sections:

        - The ``DEFGATE`` statements that define the meaning of the newly introduced "noisy" gate
          names.
        - The ``PRAGMA ADD-KRAUS`` statements to overload these noisy gates on specific qubit
          targets with their noisy implementation.
        - THe ``PRAGMA READOUT-POVM`` statements that define the noisy readout per qubit.

    :param noise_model: The assumed noise model.
    :return: A quil Program with the noise pragmas.
    """
    from pyquil.quil import Program

    p = Program()
    defgates: Set[str] = set()
    for k in noise_model.gates:

        # obtain ideal gate matrix and new, noisy name by looking it up in the NOISY_GATES dict
        try:
            ideal_gate, new_name = get_noisy_gate(k.gate, tuple(k.params))

            # if ideal version of gate has not yet been DEFGATE'd, do this
            if new_name not in defgates:
                p.defgate(new_name, ideal_gate)
                defgates.add(new_name)
        except NoisyGateUndefined:
            print(
                "WARNING: Could not find ideal gate definition for gate {}".format(k.gate),
                file=sys.stderr,
            )
            new_name = k.gate

        # define noisy version of gate on specific targets
        p.define_noisy_gate(new_name, k.targets, k.kraus_ops)

    # define noisy readouts
    for q, ap in noise_model.assignment_probs.items():
        p.define_noisy_readout(q, p00=ap[0, 0], p11=ap[1, 1])
    return p


def apply_noise_model(prog: "Program", noise_model: NoiseModel) -> "Program":
    """
    Apply a noise model to a program and generated a 'noisy-fied' version of the program.

    :param prog: A Quil Program object.
    :param noise_model: A NoiseModel, either generated from an ISA or
        from a simple decoherence model.
    :return: A new program translated to a noisy gateset and with noisy readout as described by the
        noisemodel.
    """
    new_prog = _noise_model_program_header(noise_model)
    for i in prog:
        if isinstance(i, Gate) and noise_model.gates:
            try:
                _, new_name = get_noisy_gate(i.name, tuple(i.params))
                new_prog += Gate(new_name, [], i.qubits)
            except NoisyGateUndefined:
                new_prog += i
        else:
            new_prog += i
    return new_prog


def add_decoherence_noise(
    prog: "Program",
    T1: Union[Dict[int, float], float] = 30e-6,
    T2: Union[Dict[int, float], float] = 30e-6,
    gate_time_1q: float = 50e-9,
    gate_time_2q: float = 150e-09,
    ro_fidelity: Union[Dict[int, float], float] = 0.95,
) -> "Program":
    """
    Add generic damping and dephasing noise to a program.

    This high-level function is provided as a convenience to investigate the effects of a
    generic noise model on a program. For more fine-grained control, please investigate
    the other methods available in the ``pyquil.noise`` module.

    In an attempt to closely model the QPU, noisy versions of RX(+-pi/2) and CZ are provided;
    I and parametric RZ are noiseless, and other gates are not allowed. To use this function,
    you need to compile your program to this native gate set.

    The default noise parameters

    - T1 = 30 us
    - T2 = 30 us
    - 1q gate time = 50 ns
    - 2q gate time = 150 ns

    are currently typical for near-term devices.

    This function will define new gates and add Kraus noise to these gates. It will translate
    the input program to use the noisy version of the gates.

    :param prog: A pyquil program consisting of I, RZ, CZ, and RX(+-pi/2) instructions
    :param T1: The T1 amplitude damping time either globally or in a
        dictionary indexed by qubit id. By default, this is 30 us.
    :param T2: The T2 dephasing time either globally or in a
        dictionary indexed by qubit id. By default, this is also 30 us.
    :param gate_time_1q: The duration of the one-qubit gates, namely RX(+pi/2) and RX(-pi/2).
        By default, this is 50 ns.
    :param gate_time_2q: The duration of the two-qubit gates, namely CZ.
        By default, this is 150 ns.
    :param ro_fidelity: The readout assignment fidelity
        :math:`F = (p(0|0) + p(1|1))/2` either globally or in a dictionary indexed by qubit id.
    :return: A new program with noisy operators.
    """
    gates = _get_program_gates(prog)
    noise_model = _decoherence_noise_model(
        gates,
        T1=T1,
        T2=T2,
        gate_time_1q=gate_time_1q,
        gate_time_2q=gate_time_2q,
        ro_fidelity=ro_fidelity,
    )
    return apply_noise_model(prog, noise_model)


def _bitstring_probs_by_qubit(p: np.ndarray) -> np.ndarray:
    """
    Ensure that an array ``p`` with bitstring probabilities has a separate axis for each qubit such
    that ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.

    This should not allocate much memory if ``p`` is already in ``C``-contiguous order (row-major).

    :param p: An array that enumerates bitstring probabilities. When flattened out
        ``p = [p_00...0, p_00...1, ...,p_11...1]``. The total number of elements must therefore be a
        power of 2.
    :return: A reshaped view of ``p`` with a separate length-2 axis for each bit.
    """
    p = np.asarray(p, order="C")
    num_qubits = int(round(np.log2(p.size)))
    return p.reshape((2,) * num_qubits)


def estimate_bitstring_probs(results: np.ndarray) -> np.ndarray:
    """
    Given an array of single shot results estimate the probability distribution over all bitstrings.

    :param results: A 2d array where the outer axis iterates over shots
        and the inner axis over bits.
    :return: An array with as many axes as there are qubit and normalized such that it sums to one.
        ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.
    """
    nshots, nq = np.shape(results)
    outcomes = np.array([int("".join(map(str, r)), 2) for r in results])
    probs = np.histogram(outcomes, bins=np.arange(-0.5, 2 ** nq, 1))[0] / float(nshots)
    return _bitstring_probs_by_qubit(probs)


_CHARS = "klmnopqrstuvwxyzabcdefgh0123456789"


def _apply_local_transforms(p: np.ndarray, ts: Iterable[np.ndarray]) -> np.ndarray:
    """
    Given a 2d array of single shot results (outer axis iterates over shots, inner axis over bits)
    and a list of assignment probability matrices (one for each bit in the readout, ordered like
    the inner axis of results) apply local 2x2 matrices to each bit index.

    :param p: An array that enumerates a function indexed by
        bitstrings::

            f(ijk...) = p[i,j,k,...]

    :param ts: A sequence of 2x2 transform-matrices, one for each bit.
    :return: ``p_transformed`` an array with as many dimensions as there are bits with the result of
        contracting p along each axis by the corresponding bit transformation::

            p_transformed[ijk...] = f'(ijk...) = sum_lmn... ts[0][il] ts[1][jm] ts[2][kn] f(lmn...)
    """
    p_corrected = _bitstring_probs_by_qubit(p)
    nq = p_corrected.ndim
    for idx, trafo_idx in enumerate(ts):

        # this contraction pattern looks like
        # 'ij,abcd...jklm...->abcd...iklm...' so it properly applies a "local"
        # transformation to a single tensor-index without changing the order of
        # indices
        einsum_pat = (
            "ij," + _CHARS[:idx] + "j" + _CHARS[idx : nq - 1] + "->" + _CHARS[:idx] + "i" + _CHARS[idx : nq - 1]
        )
        p_corrected = np.einsum(einsum_pat, trafo_idx, p_corrected)

    return p_corrected


def corrupt_bitstring_probs(p: np.ndarray, assignment_probabilities: List[np.ndarray]) -> np.ndarray:
    """
    Given a 2d array of true bitstring probabilities (outer axis iterates over shots, inner axis
    over bits) and a list of assignment probability matrices (one for each bit in the readout,
    ordered like the inner axis of results) compute the corrupted probabilities.

    :param p: An array that enumerates bitstring probabilities. When
        flattened out ``p = [p_00...0, p_00...1, ...,p_11...1]``. The total number of elements must
        therefore be a power of 2. The canonical shape has a separate axis for each qubit, such that
        ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.
    :param assignment_probabilities: A list of assignment probability matrices
        per qubit. Each assignment probability matrix is expected to be of the form::

            [[p00 p01]
             [p10 p11]]

    :return: ``p_corrected`` an array with as many dimensions as there are qubits that contains
        the noisy-readout-corrected estimated probabilities for each measured bitstring, i.e.,
        ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.
    """
    return _apply_local_transforms(p, assignment_probabilities)


def correct_bitstring_probs(p: np.ndarray, assignment_probabilities: List[np.ndarray]) -> np.ndarray:
    """
    Given a 2d array of corrupted bitstring probabilities (outer axis iterates over shots, inner
    axis over bits) and a list of assignment probability matrices (one for each bit in the readout)
    compute the corrected probabilities.

    :param p: An array that enumerates bitstring probabilities. When
        flattened out ``p = [p_00...0, p_00...1, ...,p_11...1]``. The total number of elements must
        therefore be a power of 2. The canonical shape has a separate axis for each qubit, such that
        ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.
    :param assignment_probabilities: A list of assignment probability matrices
        per qubit. Each assignment probability matrix is expected to be of the form::

            [[p00 p01]
             [p10 p11]]

    :return: ``p_corrected`` an array with as many dimensions as there are qubits that contains
        the noisy-readout-corrected estimated probabilities for each measured bitstring, i.e.,
        ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.
    """
    return _apply_local_transforms(p, (np.linalg.inv(ap) for ap in assignment_probabilities))


def bitstring_probs_to_z_moments(p: np.ndarray) -> np.ndarray:
    """
    Convert between bitstring probabilities and joint Z moment expectations.

    :param p: An array that enumerates bitstring probabilities. When
        flattened out ``p = [p_00...0, p_00...1, ...,p_11...1]``. The total number of elements must
        therefore be a power of 2. The canonical shape has a separate axis for each qubit, such that
        ``p[i,j,...,k]`` gives the estimated probability of bitstring ``ij...k``.
    :return: ``z_moments``, an np.array with one length-2 axis per qubit which contains the
        expectations of all monomials in ``{I, Z_0, Z_1, ..., Z_{n-1}}``. The expectations of each
        monomial can be accessed via::

            <Z_0^j_0 Z_1^j_1 ... Z_m^j_m> = z_moments[j_0,j_1,...,j_m]
    """
    zmat = np.array([[1, 1], [1, -1]])
    return _apply_local_transforms(p, (zmat for _ in range(p.ndim)))


def estimate_assignment_probs(
    q: int,
    trials: int,
    cxn: Union["QVMConnection", "QPUConnection"],
    p0: Optional["Program"] = None,
) -> np.ndarray:
    """
    Estimate the readout assignment probabilities for a given qubit ``q``.
    The returned matrix is of the form::

            [[p00 p01]
             [p10 p11]]

    :param q: The index of the qubit.
    :param trials: The number of samples for each state preparation.
    :param cxn: The quantum abstract machine to sample from.
    :param p0: A header program to prepend to the state preparation programs.
    :return: The assignment probability matrix
    """
    from pyquil.quil import Program

    if p0 is None:  # pragma no coverage
        p0 = Program()
    results_i = np.sum(cxn.run(p0 + Program(I(q), MEASURE(q, MemoryReference("ro", 0))), [0], trials))
    results_x = np.sum(cxn.run(p0 + Program(X(q), MEASURE(q, MemoryReference("ro", 0))), [0], trials))

    p00 = 1.0 - results_i / float(trials)
    p11 = results_x / float(trials)
    return np.array([[p00, 1 - p11], [1 - p00, p11]])



from pyquil.simulation._numpy import (
    NumpyWavefunctionSimulator,
    get_measure_probabilities,
    targeted_einsum,
    targeted_tensordot,
)
from pyquil.simulation._reference import (
    ReferenceDensitySimulator,
    ReferenceWavefunctionSimulator,
    zero_state_matrix,
)



from qcs_api_client.models import InstructionSetArchitecture
from qcs_api_client.operations.sync import get_instruction_set_architecture
from pyquil.external.rpcq import CompilerISA
from pyquil.device.transformers import qcs_isa_to_compiler_isa, qcs_isa_to_graph
from pyquil.device import AbstractDevice
from pyquil.noise import NoiseModel
from pyquil.api import Client
import networkx as nx
from typing import List, Optional


class QCSDevice(AbstractDevice):
    """
    An AbstractDevice initialized with an ``InstructionSetArchitecture`` returned
    from the QCS API. Notably, this class is able to serialize a ``CompilerISA`` based
    on the architecture instructions.
    """

    quantum_processor_id: str
    _isa: InstructionSetArchitecture
    noise_model: Optional[NoiseModel]

    def __init__(
        self,
        quantum_processor_id: str,
        isa: InstructionSetArchitecture,
        noise_model: Optional[NoiseModel] = None,
    ):
        """
        Initialize a new QCSDevice.

        :param quantum_processor_id: The id of the quantum processor.
        :param isa: The QCS API ``InstructionSetArchitecture``.
        :param noise_model: An optional ``NoiseModel`` for configuring a noisy device on the ``QVM``.
        """

        self.quantum_processor_id = quantum_processor_id
        self._isa = isa
        self.noise_model = noise_model

    def qubits(self) -> List[int]:
        return sorted(node.node_id for node in self._isa.architecture.nodes)

    def qubit_topology(self) -> nx.Graph:
        return qcs_isa_to_graph(self._isa)

    def to_compiler_isa(self) -> CompilerISA:
        return qcs_isa_to_compiler_isa(self._isa)

    def __str__(self) -> str:
        return "<QCSDevice {}>".format(self.quantum_processor_id)

    def __repr__(self) -> str:
        return str(self)


def get_qcs_device(client: Client, quantum_processor_id: str) -> QCSDevice:
    isa = client.qcs_request(get_instruction_set_architecture, quantum_processor_id=quantum_processor_id)

    return QCSDevice(quantum_processor_id=quantum_processor_id, isa=isa)



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module containing the Wavefunction object and methods for working with wavefunctions.
"""
import itertools
import warnings
from typing import Dict, Iterator, List, Optional, Sequence, cast

import numpy as np

OCTETS_PER_DOUBLE_FLOAT = 8
OCTETS_PER_COMPLEX_DOUBLE = 2 * OCTETS_PER_DOUBLE_FLOAT


class Wavefunction(object):
    """
    Encapsulate a wavefunction representing a quantum state
    as returned by :py:class:`~pyquil.api.WavefunctionSimulator`.

    .. note::

        The elements of the wavefunction are ordered by bitstring. E.g., for two qubits the order
        is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
        i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
        See also :ref:`the related documentation section in the WavefunctionSimulator Overview
        <basis_ordering>`.
    """

    def __init__(self, amplitude_vector: np.ndarray):
        """
        Initializes a wavefunction

        :param amplitude_vector: A numpy array of complex amplitudes
        """
        if len(amplitude_vector) == 0 or len(amplitude_vector) & (len(amplitude_vector) - 1) != 0:
            raise TypeError("Amplitude vector must have a length that is a power of two")

        self.amplitudes = np.asarray(amplitude_vector)
        sumprob = np.sum(self.probabilities())
        if not np.isclose(sumprob, 1.0):
            raise ValueError(
                "The wavefunction is not normalized. " "The probabilities sum to {} instead of 1".format(sumprob)
            )

    @staticmethod
    def ground(qubit_num: int) -> "Wavefunction":
        warnings.warn("ground() has been deprecated in favor of zeros()", stacklevel=2)
        return Wavefunction.zeros(qubit_num)

    @staticmethod
    def zeros(qubit_num: int) -> "Wavefunction":
        """
        Constructs the groundstate wavefunction for a given number of qubits.

        :param qubit_num:
        :return: A Wavefunction in the ground state
        """
        amplitude_vector = np.zeros(2 ** qubit_num)
        amplitude_vector[0] = 1.0
        return Wavefunction(amplitude_vector)

    @staticmethod
    def from_bit_packed_string(coef_string: bytes) -> "Wavefunction":
        """
        From a bit packed string, unpacks to get the wavefunction
        :param coef_string:
        """
        num_cfloat = len(coef_string) // OCTETS_PER_COMPLEX_DOUBLE
        amplitude_vector = np.ndarray(shape=(num_cfloat,), buffer=coef_string, dtype=">c16")
        return Wavefunction(amplitude_vector)

    def __len__(self) -> int:
        return len(self.amplitudes).bit_length() - 1

    def __iter__(self) -> Iterator[complex]:
        return cast(Iterator[complex], self.amplitudes.__iter__())

    def __getitem__(self, index: int) -> complex:
        return cast(complex, self.amplitudes[index])

    def __setitem__(self, key: int, value: complex) -> None:
        self.amplitudes[key] = value

    def __str__(self) -> str:
        return self.pretty_print(decimal_digits=10)

    def probabilities(self) -> np.ndarray:
        """Returns an array of probabilities in lexicographical order"""
        return np.abs(self.amplitudes) ** 2  # type: ignore

    def get_outcome_probs(self) -> Dict[str, float]:
        """
        Parses a wavefunction (array of complex amplitudes) and returns a dictionary of
        outcomes and associated probabilities.

        :return: A dict with outcomes as keys and probabilities as values.
        :rtype: dict
        """
        outcome_dict = {}
        qubit_num = len(self)
        for index, amplitude in enumerate(self.amplitudes):
            outcome = get_bitstring_from_index(index, qubit_num)
            outcome_dict[outcome] = abs(amplitude) ** 2
        return outcome_dict

    def pretty_print_probabilities(self, decimal_digits: int = 2) -> Dict[str, float]:
        """
        TODO: This doesn't seem like it is named correctly...

        Prints outcome probabilities, ignoring all outcomes with approximately zero probabilities
        (up to a certain number of decimal digits) and rounding the probabilities to decimal_digits.

        :param int decimal_digits: The number of digits to truncate to.
        :return: A dict with outcomes as keys and probabilities as values.
        """
        outcome_dict = {}
        qubit_num = len(self)
        for index, amplitude in enumerate(self.amplitudes):
            outcome = get_bitstring_from_index(index, qubit_num)
            prob = round(abs(amplitude) ** 2, decimal_digits)
            if prob != 0.0:
                outcome_dict[outcome] = prob
        return outcome_dict

    def pretty_print(self, decimal_digits: int = 2) -> str:
        """
        Returns a string repr of the wavefunction, ignoring all outcomes with approximately zero
        amplitude (up to a certain number of decimal digits) and rounding the amplitudes to
        decimal_digits.

        :param int decimal_digits: The number of digits to truncate to.
        :return: A string representation of the wavefunction.
        """
        outcome_dict = {}
        qubit_num = len(self)
        pp_string = ""
        for index, amplitude in enumerate(self.amplitudes):
            outcome = get_bitstring_from_index(index, qubit_num)
            amplitude = round(amplitude.real, decimal_digits) + round(amplitude.imag, decimal_digits) * 1.0j
            if amplitude != 0.0:
                outcome_dict[outcome] = amplitude
                pp_string += str(amplitude) + "|{}> + ".format(outcome)
        if len(pp_string) >= 3:
            pp_string = pp_string[:-3]  # remove the dangling + if it is there
        return pp_string

    def plot(self, qubit_subset: Optional[Sequence[int]] = None) -> None:
        """
        TODO: calling this will error because of matplotlib

        Plots a bar chart with bitstring on the x axis and probability on the y axis.

        :param qubit_subset: Optional parameter used for plotting a subset of the Hilbert space.
        """
        import matplotlib.pyplot as plt

        prob_dict = self.get_outcome_probs()
        if qubit_subset:
            sub_dict = {}
            qubit_num = len(self)
            for i in qubit_subset:
                if i > (2 ** qubit_num - 1):
                    raise IndexError("Index {} too large for {} qubits.".format(i, qubit_num))
                else:
                    sub_dict[get_bitstring_from_index(i, qubit_num)] = prob_dict[get_bitstring_from_index(i, qubit_num)]
            prob_dict = sub_dict
        plt.bar(range(len(prob_dict)), prob_dict.values(), align="center", color="#6CAFB7")
        plt.xticks(range(len(prob_dict)), prob_dict.keys())
        plt.show()

    def sample_bitstrings(self, n_samples: int) -> np.ndarray:
        """
        Sample bitstrings from the distribution defined by the wavefunction.

        :param n_samples: The number of bitstrings to sample
        :return: An array of shape (n_samples, n_qubits)
        """
        possible_bitstrings = np.array(list(itertools.product((0, 1), repeat=len(self))))
        inds = np.random.choice(2 ** len(self), n_samples, p=self.probabilities())
        bitstrings: np.ndarray = possible_bitstrings[inds, :]
        return bitstrings


def get_bitstring_from_index(index: int, qubit_num: int) -> str:
    """
    Returns the bitstring in lexical order that corresponds to the given index in 0 to 2^(qubit_num)
    :param int index:
    :param int qubit_num:
    :return: the bitstring
    :rtype: str
    """
    if index > (2 ** qubit_num - 1):
        raise IndexError("Index {} too large for {} qubits.".format(index, qubit_num))
    return bin(index)[2:].rjust(qubit_num, "0")


def _octet_bits(o: int) -> List[int]:
    """
    Get the bits of an octet.

    :param o: The octets.
    :return: The bits as a list in LSB-to-MSB order.
    """
    if not isinstance(o, int):
        raise TypeError("o should be an int")
    if not (0 <= o <= 255):
        raise ValueError("o should be between 0 and 255 inclusive")
    bits = [0] * 8
    for i in range(8):
        if 1 == o & 1:
            bits[i] = 1
        o = o >> 1
    return bits



from pyquil.device.transformers.qcs_isa_to_compiler_isa import (
    qcs_isa_to_compiler_isa,
    QCSISAParseError,
)
from pyquil.device.transformers.qcs_isa_to_graph import qcs_isa_to_graph
from pyquil.device.transformers.compiler_isa_to_graph import compiler_isa_to_graph
from pyquil.device.transformers.graph_to_compiler_isa import graph_to_compiler_isa, GraphGateError



from pyquil.device._base import AbstractDevice
from typing import List
from pyquil.device.transformers import compiler_isa_to_graph
from pyquil.external.rpcq import CompilerISA

import networkx as nx


class CompilerDevice(AbstractDevice):
    """
    An AbstractDevice initialized with a user constructed ``CompilerISA``.
    """

    _isa: CompilerISA

    def __init__(self, isa: CompilerISA) -> None:
        self._isa = isa

    def qubit_topology(self) -> nx.Graph:
        return compiler_isa_to_graph(self._isa)

    def to_compiler_isa(self) -> CompilerISA:
        return self._isa

    def qubits(self) -> List[int]:
        return sorted([int(node_id) for node_id, node in self._isa.qubits.items()])



from typing import Dict, Any
from pyquil.external.rpcq import make_edge_id
from pyquil.device import QCSDevice
from pyquil.device.transformers import qcs_isa_to_compiler_isa
from pyquil.noise import NoiseModel
from pyquil.external.rpcq import CompilerISA
from qcs_api_client.models import InstructionSetArchitecture


def test_qcs_isa_to_compiler_isa(qcs_aspen8_isa: InstructionSetArchitecture, aspen8_compiler_isa: CompilerISA):
    """
    Test that ``qcs_isa_to_compiler_isa`` accurately transforms an ``InstructionSetArchitecture``
    to a ``CompilerISA``. The qubits and edges in the former should all be present in the latter.

    Note, this is a data driven test using fixtures defined in conftest.py. This
    comprehensively checks that ``qcs_isa_to_compiler_isa`` transforms all operators and fidelities
    accurately.
    """
    compiler_isa = qcs_isa_to_compiler_isa(qcs_aspen8_isa)

    for node in qcs_aspen8_isa.architecture.nodes:
        assert str(node.node_id) in compiler_isa.qubits

    for edge in qcs_aspen8_isa.architecture.edges:
        assert make_edge_id(edge.node_ids[0], edge.node_ids[1]) in compiler_isa.edges

    assert compiler_isa == aspen8_compiler_isa


def test_qcs_noise_model(qcs_aspen8_isa: InstructionSetArchitecture, noise_model_dict: Dict[str, Any]):
    """
    Test that ``NoiseModel.from_dict`` initializes a ``NoiseModel``, which users may, in turn,
    pass to ``QCSDevice`` for later initializing a noisy QVM.
    """

    noise_model = NoiseModel.from_dict(noise_model_dict)
    device = QCSDevice("Aspen-8", qcs_aspen8_isa, noise_model=noise_model)
    assert device.quantum_processor_id == "Aspen-8"

    assert isinstance(device.noise_model, NoiseModel)
    assert device.noise_model == noise_model



import numpy as np

from pyquil.experiment._memory import (
    merge_memory_map_lists,
    pauli_term_to_preparation_memory_map,
    pauli_term_to_measurement_memory_map,
)
from pyquil.paulis import sX, sY


def test_merge_memory_map_lists():
    mml = [
        {"a": [1.0], "b": [3.0], "c": [4.0]},
        {"a": [1.0], "b": [5.0], "c": [6.0]},
        {"a": [2.0], "b": [3.0], "c": [4.0]},
        {"a": [2.0], "b": [5.0], "c": [6.0]},
    ]
    assert mml == merge_memory_map_lists(
        [{"a": [1.0]}, {"a": [2.0]}], [{"b": [3.0], "c": [4.0]}, {"b": [5.0], "c": [6.0]}]
    )


def test_pauli_term_to_preparation_memory_map():
    assert pauli_term_to_preparation_memory_map(sX(0)) == {
        "preparation_alpha": [0.0],
        "preparation_beta": [np.pi / 2],
        "preparation_gamma": [0.0],
    }


def test_pauli_term_to_measurement_memory_map():
    assert pauli_term_to_measurement_memory_map(sY(1)) == {
        "measurement_alpha": [0.0, np.pi / 2],
        "measurement_beta": [0.0, np.pi / 2],
        "measurement_gamma": [0.0, -np.pi / 2],
    }



from qcs_api_client.models import InstructionSetArchitecture, Characteristic
from pyquil.external.rpcq import CompilerISA, add_edge, add_qubit, get_qubit, get_edge
import numpy as np
from pyquil.external.rpcq import (
    GateInfo,
    MeasureInfo,
    Supported1QGate,
    Supported2QGate,
    make_edge_id,
)
from typing import List, Union, cast, DefaultDict, Set
from collections import defaultdict


class QCSISAParseError(ValueError):
    """
    Signals an error when creating a ``CompilerISA`` due to the operators
    in the QCS ``InstructionSetArchitecture``. This may raise as a consequence
    of unsupported gates as well as missing nodes or edges.
    """

    pass


def qcs_isa_to_compiler_isa(isa: InstructionSetArchitecture) -> CompilerISA:
    device = CompilerISA()
    for node in isa.architecture.nodes:
        add_qubit(device, node.node_id)

    for edge in isa.architecture.edges:
        add_edge(device, edge.node_ids[0], edge.node_ids[1])

    qubit_operations_seen: DefaultDict[int, Set[str]] = defaultdict(set)
    edge_operations_seen: DefaultDict[str, Set[str]] = defaultdict(set)
    for operation in isa.instructions:
        for site in operation.sites:
            if operation.node_count == 1:
                if len(site.node_ids) != 1:
                    raise QCSISAParseError(
                        f"operation {operation.name} has node count 1, but " f"site has {len(site.node_ids)} node_ids"
                    )
                operation_qubit = get_qubit(device, site.node_ids[0])
                if operation_qubit is None:
                    raise QCSISAParseError(
                        f"operation {operation.name} has node {site.node_ids[0]} "
                        "but node not declared in architecture"
                    )

                if operation.name in qubit_operations_seen[operation_qubit.id]:
                    continue
                qubit_operations_seen[operation_qubit.id].add(operation.name)

                operation_qubit.gates.extend(
                    _transform_qubit_operation_to_gates(operation.name, operation_qubit.id, site.characteristics)
                )

            elif operation.node_count == 2:
                if len(site.node_ids) != 2:
                    QCSISAParseError(
                        f"operation {operation.name} has node count 2, but site " f"has {len(site.node_ids)} node_ids"
                    )

                operation_edge = get_edge(device, site.node_ids[0], site.node_ids[1])
                edge_id = make_edge_id(site.node_ids[0], site.node_ids[1])
                if operation_edge is None:
                    raise QCSISAParseError(
                        f"operation {operation.name} has site {site.node_ids}, but edge {edge_id} "
                        f"not declared in architecture"
                    )

                if operation.name in edge_operations_seen[edge_id]:
                    continue
                edge_operations_seen[edge_id].add(operation.name)

                operation_edge.gates.extend(_transform_edge_operation_to_gates(operation.name, site.characteristics))

            else:
                raise QCSISAParseError("unexpected operation node count: {}".format(operation.node_count))

    return device


PERFECT_FIDELITY = 1e0
PERFECT_DURATION = 1 / 100

_operation_names_to_compiler_fidelity_default = {
    Supported2QGate.CZ: 0.89,
    Supported2QGate.ISWAP: 0.90,
    Supported2QGate.CPHASE: 0.85,
    Supported2QGate.XY: 0.86,
    Supported1QGate.RX: 0.95,
    Supported1QGate.MEASURE: 0.90,
}

_operation_names_to_compiler_duration_default = {
    Supported2QGate.CZ: 200,
    Supported2QGate.ISWAP: 200,
    Supported2QGate.CPHASE: 200,
    Supported2QGate.XY: 200,
    Supported1QGate.RX: 50,
    Supported1QGate.MEASURE: 2000,
}


def _make_measure_gates(node_id: int, characteristics: List[Characteristic]) -> List[MeasureInfo]:
    duration = _operation_names_to_compiler_duration_default[Supported1QGate.MEASURE]
    fidelity = _operation_names_to_compiler_fidelity_default[Supported1QGate.MEASURE]
    for characteristic in characteristics:
        if characteristic.name == "fRO":
            fidelity = characteristic.value
            break

    return [
        MeasureInfo(
            operator=Supported1QGate.MEASURE,
            qubit=str(node_id),
            target="_",
            fidelity=fidelity,
            duration=duration,
        ),
        MeasureInfo(
            operator=Supported1QGate.MEASURE,
            qubit=str(node_id),
            target=None,
            fidelity=fidelity,
            duration=duration,
        ),
    ]


def _make_rx_gates(node_id: int, characteristics: List[Characteristic]) -> List[GateInfo]:
    default_duration = _operation_names_to_compiler_duration_default[Supported1QGate.RX]

    default_fidelity = _operation_names_to_compiler_fidelity_default[Supported1QGate.RX]
    fidelity = default_fidelity
    for characteristic in characteristics:
        if characteristic.name == "f1QRB":
            fidelity = characteristic.value
            break

    gates = [
        GateInfo(
            operator=Supported1QGate.RX,
            parameters=[0.0],
            arguments=[node_id],
            fidelity=PERFECT_FIDELITY,
            duration=default_duration,
        )
    ]
    for param in [np.pi, -np.pi, np.pi / 2, -np.pi / 2]:
        gates.append(
            GateInfo(
                operator=Supported1QGate.RX,
                parameters=[param],
                arguments=[node_id],
                fidelity=fidelity,
                duration=default_duration,
            )
        )
    return gates


def _make_rz_gates(node_id: int) -> List[GateInfo]:
    return [
        GateInfo(
            operator=Supported1QGate.RZ,
            parameters=["_"],
            arguments=[node_id],
            fidelity=PERFECT_FIDELITY,
            duration=PERFECT_DURATION,
        )
    ]


def _make_wildcard_1q_gates(node_id: int) -> List[GateInfo]:
    return [
        GateInfo(
            operator="_",
            parameters=["_"],
            arguments=[node_id],
            fidelity=PERFECT_FIDELITY,
            duration=PERFECT_DURATION,
        )
    ]


def _transform_qubit_operation_to_gates(
    operation_name: str,
    node_id: int,
    characteristics: List[Characteristic],
) -> List[Union[GateInfo, MeasureInfo]]:
    if operation_name == Supported1QGate.RX:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_rx_gates(node_id, characteristics))
    elif operation_name == Supported1QGate.RZ:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_rz_gates(node_id))
    elif operation_name == Supported1QGate.MEASURE:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_measure_gates(node_id, characteristics))
    elif operation_name == Supported1QGate.WILDCARD:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_wildcard_1q_gates(node_id))
    elif operation_name in {"I", "RESET"}:
        return []
    else:
        raise QCSISAParseError("Unsupported qubit operation: {}".format(operation_name))


def _make_cz_gates(characteristics: List[Characteristic]) -> List[GateInfo]:
    default_duration = _operation_names_to_compiler_duration_default[Supported2QGate.CZ]

    default_fidelity = _operation_names_to_compiler_fidelity_default[Supported2QGate.CZ]
    fidelity = default_fidelity
    for characteristic in characteristics:
        if characteristic.name == "fCZ":
            fidelity = characteristic.value
            break

    return [
        GateInfo(
            operator=Supported2QGate.CZ,
            parameters=[],
            arguments=["_", "_"],
            fidelity=fidelity,
            duration=default_duration,
        )
    ]


def _make_iswap_gates(characteristics: List[Characteristic]) -> List[GateInfo]:
    default_duration = _operation_names_to_compiler_duration_default[Supported2QGate.ISWAP]

    default_fidelity = _operation_names_to_compiler_fidelity_default[Supported2QGate.ISWAP]
    fidelity = default_fidelity
    for characteristic in characteristics:
        if characteristic.name == "fISWAP":
            fidelity = characteristic.value
            break

    return [
        GateInfo(
            operator=Supported2QGate.ISWAP,
            parameters=[],
            arguments=["_", "_"],
            fidelity=fidelity,
            duration=default_duration,
        )
    ]


def _make_cphase_gates(characteristics: List[Characteristic]) -> List[GateInfo]:
    default_duration = _operation_names_to_compiler_duration_default[Supported2QGate.CPHASE]

    default_fidelity = _operation_names_to_compiler_fidelity_default[Supported2QGate.CPHASE]
    fidelity = default_fidelity
    for characteristic in characteristics:
        if characteristic.name == "fCPHASE":
            fidelity = characteristic.value
            break

    return [
        GateInfo(
            operator=Supported2QGate.ISWAP,
            parameters=["theta"],
            arguments=["_", "_"],
            fidelity=fidelity,
            duration=default_duration,
        )
    ]


def _make_xy_gates(characteristics: List[Characteristic]) -> List[GateInfo]:
    default_duration = _operation_names_to_compiler_duration_default[Supported2QGate.XY]

    default_fidelity = _operation_names_to_compiler_fidelity_default[Supported2QGate.XY]
    fidelity = default_fidelity
    for characteristic in characteristics:
        if characteristic.name == "fXY":
            fidelity = characteristic.value
            break

    return [
        GateInfo(
            operator=Supported2QGate.XY,
            parameters=["theta"],
            arguments=["_", "_"],
            fidelity=fidelity,
            duration=default_duration,
        )
    ]


def _make_wildcard_2q_gates() -> List[GateInfo]:
    return [
        GateInfo(
            operator="_",
            parameters=["_"],
            arguments=["_", "_"],
            fidelity=PERFECT_FIDELITY,
            duration=PERFECT_DURATION,
        )
    ]


def _transform_edge_operation_to_gates(
    operation_name: str,
    characteristics: List[Characteristic],
) -> List[GateInfo]:
    if operation_name == Supported2QGate.CZ:
        return _make_cz_gates(characteristics)
    elif operation_name == Supported2QGate.ISWAP:
        return _make_iswap_gates(characteristics)
    elif operation_name == Supported2QGate.CPHASE:
        return _make_cphase_gates(characteristics)
    elif operation_name == Supported2QGate.XY:
        return _make_xy_gates(characteristics)
    elif operation_name == Supported2QGate.WILDCARD:
        return _make_wildcard_2q_gates()
    else:
        raise QCSISAParseError("Unsupported edge operation: {}".format(operation_name))



import pytest
from pyquil.gates import (
    CNOT,
    CPHASE,
    CPHASE00,
    CPHASE01,
    CPHASE10,
    CZ,
    H,
    I,
    ISWAP,
    PHASE,
    PSWAP,
    RX,
    RY,
    RZ,
    S,
    SWAP,
    T,
    X,
    Y,
    Z,
)
from pyquil.quilbase import _strip_modifiers


@pytest.fixture(params=[I, X, Y, Z, H, S, T])
def oneq_gate(request):
    return request.param


@pytest.fixture(params=[PHASE, RX, RY, RZ])
def param_oneq_gate(request):
    return request.param


@pytest.fixture(params=[CZ, CNOT, SWAP, ISWAP])
def twoq_gate(request):
    return request.param


@pytest.fixture(params=[CPHASE, CPHASE00, CPHASE01, CPHASE10, PSWAP])
def param_twoq_gate(request):
    return request.param


def test_oneq_gate(oneq_gate):
    g = oneq_gate(234)
    assert g.out() == "{} 234".format(g.name)

    func_name = oneq_gate.__name__
    assert g.name == func_name


def test_oneq_gate_kwarg(oneq_gate):
    g = oneq_gate(qubit=234)
    assert g.out() == "{} 234".format(g.name)


def test_param_oneq_gate(param_oneq_gate):
    g = param_oneq_gate(0.2, 234)
    assert g.out() == "{}(0.2) 234".format(g.name)

    func_name = param_oneq_gate.__name__
    assert g.name == func_name


def test_param_oneq_gate_kwarg(param_oneq_gate):
    g = param_oneq_gate(angle=0.2, qubit=234)
    assert g.out() == "{}(0.2) 234".format(g.name)


def test_twoq_gate(twoq_gate):
    g = twoq_gate(234, 567)
    assert g.out() == "{} 234 567".format(g.name)

    func_name = twoq_gate.__name__
    assert g.name == func_name


def test_twoq_gate_kwarg(twoq_gate):
    func_name = twoq_gate.__name__
    if func_name.startswith("C"):
        qubits = {"control": 234, "target": 567}
    elif "SWAP" in func_name:
        qubits = {"q1": 234, "q2": 567}
    else:
        raise ValueError()

    g = twoq_gate(**qubits)
    assert g.out() == "{} 234 567".format(g.name)


def test_param_twoq_gate(param_twoq_gate):
    g = param_twoq_gate(0.2, 234, 567)
    assert g.out() == "{}(0.2) 234 567".format(g.name)

    func_name = param_twoq_gate.__name__
    assert g.name == func_name


def test_param_twoq_gate_kwarg(param_twoq_gate):
    func_name = param_twoq_gate.__name__
    if func_name.startswith("C"):
        qubits = {"control": 234, "target": 567}
    elif "SWAP" in func_name:
        qubits = {"q1": 234, "q2": 567}
    else:
        raise ValueError()
    g = param_twoq_gate(angle=0.2, **qubits)
    assert g.out() == "{}(0.2) 234 567".format(g.name)


def test_controlled_gate():
    g = X(0).controlled(1)
    assert g.out() == "CONTROLLED X 1 0"
    g = X(0).controlled(1).controlled(2)
    assert g.out() == "CONTROLLED CONTROLLED X 2 1 0"
    g = X(0).controlled([1, 2])
    assert g.out() == "CONTROLLED CONTROLLED X 2 1 0"
    # for backwards compatibility
    g = X(0).controlled(control_qubit=1)
    assert g.out() == "CONTROLLED X 1 0"
    g = X(0).controlled([])
    assert g.out() == "X 0"


def test_dagger_gate():
    g = X(0).dagger()
    assert g.out() == "DAGGER X 0"
    # This will be compiled away by quilc
    g = X(0).dagger().dagger() == "DAGGER DAGGER X 0"


def test_forked_gate():
    g = RX(0.0, 0).forked(1, [1.0])
    assert g.out() == "FORKED RX(0,1.0) 1 0"
    g = RX(0.0, 0).forked(1, [1.0]).forked(2, [2.0, 3.0])
    assert g.out() == "FORKED FORKED RX(0,1.0,2.0,3.0) 2 1 0"


def test_dagger_controlled_gate():
    g = X(0).dagger().controlled(1)
    assert g.out() == "CONTROLLED DAGGER X 1 0"
    g = X(0).controlled(1).dagger()
    assert g.out() == "DAGGER CONTROLLED X 1 0"
    g = X(0).controlled(1).dagger().controlled(2)
    assert g.out() == "CONTROLLED DAGGER CONTROLLED X 2 1 0"


def test_mixed_gate_modifiers():
    g = RX(0.1, 3).forked(2, [0.2]).controlled(1).dagger().forked(0, [0.3, 0.4])
    assert g.out() == "FORKED DAGGER CONTROLLED FORKED RX(0.1,0.2,0.3,0.4) 0 1 2 3"


def test_strip_gate_modifiers():
    g0 = RX(0.1, 3)
    g1 = RX(0.1, 3).forked(2, [0.2]).controlled(1)
    g2 = RX(0.1, 3).forked(2, [0.2]).controlled(1).dagger()

    assert _strip_modifiers(g1) == g0
    assert _strip_modifiers(g2) == g0
    assert _strip_modifiers(g2, 3) == g0
    assert _strip_modifiers(g2, 1) == g1



import functools
from operator import mul

import numpy as np
import pytest

from pyquil.experiment._setting import (
    ExperimentSetting,
    SIC0,
    SIC1,
    SIC2,
    SIC3,
    TensorProductState,
    minusX,
    minusY,
    minusZ,
    plusX,
    plusY,
    plusZ,
    zeros_state,
)
from pyquil.paulis import sI, sX, sY, sZ


def _generate_random_states(n_qubits, n_terms):
    oneq_states = [SIC0, SIC1, SIC2, SIC3, plusX, minusX, plusY, minusY, plusZ, minusZ]
    all_s_inds = np.random.randint(len(oneq_states), size=(n_terms, n_qubits))
    states = []
    for s_inds in all_s_inds:
        state = functools.reduce(mul, (oneq_states[pi](i) for i, pi in enumerate(s_inds)), TensorProductState([]))
        states += [state]
    return states


def _generate_random_paulis(n_qubits, n_terms):
    paulis = [sI, sX, sY, sZ]
    all_op_inds = np.random.randint(len(paulis), size=(n_terms, n_qubits))
    operators = []
    for op_inds in all_op_inds:
        op = functools.reduce(mul, (paulis[pi](i) for i, pi in enumerate(op_inds)), sI(0))
        op *= np.random.uniform(-1, 1)
        operators += [op]
    return operators


def test_experiment_setting():
    in_states = _generate_random_states(n_qubits=4, n_terms=7)
    out_ops = _generate_random_paulis(n_qubits=4, n_terms=7)
    for ist, oop in zip(in_states, out_ops):
        expt = ExperimentSetting(ist, oop)
        assert str(expt) == expt.serializable()
        expt2 = ExperimentSetting.from_str(str(expt))
        assert expt == expt2
        assert expt2.in_state == ist
        assert expt2.out_operator == oop


@pytest.mark.filterwarnings("ignore:ExperimentSetting")
def test_setting_no_in_back_compat():
    out_ops = _generate_random_paulis(n_qubits=4, n_terms=7)
    for oop in out_ops:
        expt = ExperimentSetting(TensorProductState(), oop)
        expt2 = ExperimentSetting.from_str(str(expt))
        assert expt == expt2
        assert expt2.in_operator == sI()
        assert expt2.out_operator == oop


@pytest.mark.filterwarnings("ignore:ExperimentSetting")
def test_setting_no_in():
    out_ops = _generate_random_paulis(n_qubits=4, n_terms=7)
    for oop in out_ops:
        expt = ExperimentSetting(zeros_state(oop.get_qubits()), oop)
        expt2 = ExperimentSetting.from_str(str(expt))
        assert expt == expt2
        assert expt2.in_operator == functools.reduce(mul, [sZ(q) for q in oop.get_qubits()], sI())
        assert expt2.out_operator == oop



import numpy as np
from unittest.mock import Mock
import pytest

from pyquil.noise import (
    pauli_kraus_map,
    damping_kraus_map,
    dephasing_kraus_map,
    tensor_kraus_maps,
    _get_program_gates,
    _decoherence_noise_model,
    add_decoherence_noise,
    combine_kraus_maps,
    damping_after_dephasing,
    INFINITY,
    apply_noise_model,
    _noise_model_program_header,
    KrausModel,
    NoiseModel,
    corrupt_bitstring_probs,
    correct_bitstring_probs,
    estimate_bitstring_probs,
    bitstring_probs_to_z_moments,
    estimate_assignment_probs,
    NO_NOISE,
)
from pyquil.gates import RZ, RX, I, CZ
from collections import OrderedDict

from pyquil.quil import Pragma, Program
from pyquil.quilbase import DefGate, Gate
from pyquil.api import QVMConnection


def test_pauli_kraus_map():
    probabilities = [0.1, 0.2, 0.3, 0.4]
    k1, k2, k3, k4 = pauli_kraus_map(probabilities)
    assert np.allclose(k1, np.sqrt(0.1) * np.eye(2), atol=1 * 10 ** -8)
    assert np.allclose(k2, np.sqrt(0.2) * np.array([[0, 1.0], [1.0, 0]]), atol=1 * 10 ** -8)
    assert np.allclose(k3, np.sqrt(0.3) * np.array([[0, -1.0j], [1.0j, 0]]), atol=1 * 10 ** -8)
    assert np.allclose(k4, np.sqrt(0.4) * np.array([[1, 0], [0, -1]]), atol=1 * 10 ** -8)

    two_q_pauli_kmaps = pauli_kraus_map(np.kron(probabilities, list(reversed(probabilities))))
    q1_pauli_kmaps = [k1, k2, k3, k4]
    q2_pauli_kmaps = pauli_kraus_map(list(reversed(probabilities)))
    tensor_kmaps = tensor_kraus_maps(q1_pauli_kmaps, q2_pauli_kmaps)
    assert np.allclose(two_q_pauli_kmaps, tensor_kmaps)


def test_damping_kraus_map():
    p = 0.05
    k1, k2 = damping_kraus_map(p=p)
    assert k1[1, 1] == np.sqrt(1 - p)
    assert k2[0, 1] == np.sqrt(p)


def test_dephasing_kraus_map():
    p = 0.05
    k1, k2 = dephasing_kraus_map(p=p)
    np.testing.assert_allclose(np.diag(k1), [np.sqrt(1 - p)] * 2)
    np.testing.assert_allclose(np.abs(np.diag(k2)), [np.sqrt(p)] * 2)


def test_tensor_kraus_maps():
    damping = damping_kraus_map()
    k1, k2, k3, k4 = tensor_kraus_maps(damping, damping)
    assert k1.shape == (4, 4)
    assert k2.shape == (4, 4)
    assert k3.shape == (4, 4)
    assert k4.shape == (4, 4)
    np.testing.assert_allclose(k1[-1, -1], 1 - 0.1)


def test_combine_kraus_maps():
    damping = damping_kraus_map()
    dephasing = dephasing_kraus_map()
    k1, k2, k3, k4 = combine_kraus_maps(damping, dephasing)
    assert k1.shape == (2, 2)
    assert k2.shape == (2, 2)
    assert k3.shape == (2, 2)
    assert k4.shape == (2, 2)


def test_damping_after_dephasing():
    gate_time = 1
    T1 = 10
    T2 = 3
    kraus_map = damping_after_dephasing(T1, T2, gate_time)

    # Density matrix for the |+> state
    rho = 0.5 * np.ones((2, 2))

    # See Eq. 7.144 of Nielsen and Chuang
    target_rho = [
        [1 - 0.5 * np.exp(-gate_time / T1), 0.5 * np.exp(-gate_time / T2)],
        [0.5 * np.exp(-gate_time / T2), 0.5 * np.exp(-gate_time / T1)],
    ]

    noisy_rho = np.zeros((2, 2))
    for op in kraus_map:
        noisy_rho += op @ rho @ (op.T.conj())

    np.testing.assert_allclose(noisy_rho, target_rho)


def test_noise_helpers():
    gates = RX(np.pi / 2, 0), RX(-np.pi / 2, 1), I(1), CZ(0, 1)
    prog = Program(*gates)
    inferred_gates = _get_program_gates(prog)
    assert set(inferred_gates) == set(gates)


def test_decoherence_noise():
    prog = Program(RX(np.pi / 2, 0), CZ(0, 1), RZ(np.pi, 0))
    gates = _get_program_gates(prog)
    m1 = _decoherence_noise_model(gates, T1=INFINITY, T2=INFINITY, ro_fidelity=1.0)

    # with no readout error, assignment_probs = identity matrix
    assert np.allclose(m1.assignment_probs[0], np.eye(2))
    assert np.allclose(m1.assignment_probs[1], np.eye(2))
    for g in m1.gates:
        # with infinite coherence time all kraus maps should only have a single, unitary kraus op
        assert len(g.kraus_ops) == 1
        (k0,) = g.kraus_ops
        # check unitarity
        k0dk0 = k0.dot(k0.conjugate().transpose())
        assert np.allclose(k0dk0, np.eye(k0dk0.shape[0]))

    # verify that selective (by qubit) dephasing and readout infidelity is working
    m2 = _decoherence_noise_model(gates, T1=INFINITY, T2={0: 30e-6}, ro_fidelity={0: 0.95, 1: 1.0})
    assert np.allclose(m2.assignment_probs[0], [[0.95, 0.05], [0.05, 0.95]])
    assert np.allclose(m2.assignment_probs[1], np.eye(2))
    for g in m2.gates:
        if 0 in g.targets:
            # single dephasing (no damping) channel on qc 0, no noise on qc1 -> 2 Kraus ops
            assert len(g.kraus_ops) == 2
        else:
            assert len(g.kraus_ops) == 1

    # verify that combined T1 and T2 will lead to 4 outcome Kraus map.
    m3 = _decoherence_noise_model(gates, T1={0: 30e-6}, T2={0: 30e-6})
    for g in m3.gates:
        if 0 in g.targets:
            # damping (implies dephasing) channel on qc 0, no noise on qc1 -> 4 Kraus ops
            assert len(g.kraus_ops) == 4
        else:
            assert len(g.kraus_ops) == 1

    # verify that gate names are translated
    new_prog = apply_noise_model(prog, m3)

    # check that headers have been embedded
    headers = _noise_model_program_header(m3)
    assert all(
        (isinstance(i, Pragma) and i.command in ["ADD-KRAUS", "READOUT-POVM"]) or isinstance(i, DefGate)
        for i in headers
    )
    assert headers.out() in new_prog.out()

    # verify that high-level add_decoherence_noise reproduces new_prog
    new_prog2 = add_decoherence_noise(prog, T1={0: 30e-6}, T2={0: 30e-6})
    assert new_prog == new_prog2


def test_kraus_model_1():
    km = KrausModel("I", (5.0,), (0, 1), [np.array([[1 + 1j]])], 1.0)
    d = km.to_dict()
    assert d == OrderedDict(
        [
            ("gate", km.gate),
            ("params", km.params),
            ("targets", (0, 1)),
            ("kraus_ops", [[[[1.0]], [[1.0]]]]),
            ("fidelity", 1.0),
        ]
    )
    assert KrausModel.from_dict(d) == km


@pytest.fixture
def kraus_model_I_dict():
    return {
        "gate": "I",
        "fidelity": 1.0,
        "kraus_ops": [[[[1.0]], [[1.0]]]],
        "targets": (0, 1),
        "params": (5.0,),
    }


def test_kraus_model_2(kraus_model_I_dict):
    km = KrausModel.from_dict(kraus_model_I_dict)
    assert km == KrausModel(
        gate=kraus_model_I_dict["gate"],
        params=kraus_model_I_dict["params"],
        targets=kraus_model_I_dict["targets"],
        kraus_ops=[KrausModel.unpack_kraus_matrix(kraus_op) for kraus_op in kraus_model_I_dict["kraus_ops"]],
        fidelity=kraus_model_I_dict["fidelity"],
    )
    d = km.to_dict()
    assert d == OrderedDict(
        [
            ("gate", km.gate),
            ("params", km.params),
            ("targets", (0, 1)),
            ("kraus_ops", [[[[1.0]], [[1.0]]]]),
            ("fidelity", 1.0),
        ]
    )


def test_noise_model_1():
    km1 = KrausModel("I", (5.0,), (0, 1), [np.array([[1 + 1j]])], 1.0)
    km2 = KrausModel("RX", (np.pi / 2,), (0,), [np.array([[1 + 1j]])], 1.0)
    nm = NoiseModel([km1, km2], {0: np.eye(2), 1: np.eye(2)})

    assert nm == NoiseModel.from_dict(nm.to_dict())
    assert nm.gates_by_name("I") == [km1]
    assert nm.gates_by_name("RX") == [km2]


@pytest.fixture
def kraus_model_RX90_dict():
    return {
        "gate": "RX",
        "fidelity": 1.0,
        "kraus_ops": [[[[1.0]], [[1.0]]]],
        "targets": (0,),
        "params": (np.pi / 2.0,),
    }


def test_noise_model_2(kraus_model_I_dict, kraus_model_RX90_dict):
    noise_model_dict = {
        "gates": [kraus_model_I_dict, kraus_model_RX90_dict],
        "assignment_probs": {"1": [[1.0, 0.0], [0.0, 1.0]], "0": [[1.0, 0.0], [0.0, 1.0]]},
    }

    nm = NoiseModel.from_dict(noise_model_dict)
    km1 = KrausModel.from_dict(kraus_model_I_dict)
    km2 = KrausModel.from_dict(kraus_model_RX90_dict)
    assert nm == NoiseModel(gates=[km1, km2], assignment_probs={0: np.eye(2), 1: np.eye(2)})
    assert nm.gates_by_name("I") == [km1]
    assert nm.gates_by_name("RX") == [km2]
    assert nm.to_dict() == noise_model_dict


def test_readout_compensation():
    np.random.seed(1234124)
    p = np.random.rand(2, 2, 2, 2, 2, 2)
    p /= p.sum()

    aps = [np.eye(2) + 0.2 * (np.random.rand(2, 2) - 1) for _ in range(p.ndim)]
    for ap in aps:
        ap.flat[ap.flat < 0] = 0.0
        ap /= ap.sum()
        assert np.alltrue(ap >= 0)

    assert np.alltrue(p >= 0)

    p_corrupted = corrupt_bitstring_probs(p, aps)
    p_restored = correct_bitstring_probs(p_corrupted, aps)
    assert np.allclose(p, p_restored)

    results = [[0, 0, 0]] * 100 + [[0, 1, 1]] * 200
    p1 = estimate_bitstring_probs(results)
    assert np.isclose(p1[0, 0, 0], 1.0 / 3.0)
    assert np.isclose(p1[0, 1, 1], 2.0 / 3.0)
    assert np.isclose(p1.sum(), 1.0)

    zm = bitstring_probs_to_z_moments(p1)
    assert np.isclose(zm[0, 0, 0], 1)
    assert np.isclose(zm[1, 0, 0], 1)
    assert np.isclose(zm[0, 1, 0], -1.0 / 3)
    assert np.isclose(zm[0, 0, 1], -1.0 / 3)
    assert np.isclose(zm[0, 1, 1], 1.0)
    assert np.isclose(zm[1, 1, 0], -1.0 / 3)
    assert np.isclose(zm[1, 0, 1], -1.0 / 3)
    assert np.isclose(zm[1, 1, 1], 1.0)


def test_estimate_assignment_probs():
    cxn = Mock(spec=QVMConnection)
    trials = 100
    p00 = 0.8
    p11 = 0.75
    cxn.run.side_effect = [
        [[0]] * int(round(p00 * trials)) + [[1]] * int(round((1 - p00) * trials)),
        [[1]] * int(round(p11 * trials)) + [[0]] * int(round((1 - p11) * trials)),
    ]
    ap_target = np.array([[p00, 1 - p11], [1 - p00, p11]])

    povm_pragma = Pragma("READOUT-POVM", (0, "({} {} {} {})".format(*ap_target.flatten())))
    ap = estimate_assignment_probs(0, trials, cxn, Program(povm_pragma))
    assert np.allclose(ap, ap_target)
    for call in cxn.run.call_args_list:
        args, kwargs = call
        prog = args[0]
        assert prog._instructions[0] == povm_pragma


def test_apply_noise_model():
    p = Program(RX(np.pi / 2, 0), RX(np.pi / 2, 1), CZ(0, 1), RX(np.pi / 2, 1))
    noise_model = _decoherence_noise_model(_get_program_gates(p))
    pnoisy = apply_noise_model(p, noise_model)
    for i in pnoisy:
        if isinstance(i, DefGate):
            pass
        elif isinstance(i, Pragma):
            assert i.command in ["ADD-KRAUS", "READOUT-POVM"]
        elif isinstance(i, Gate):
            assert i.name in NO_NOISE or not i.params


def test_apply_noise_model_perturbed_angles():
    eps = 1e-15
    p = Program(RX(np.pi / 2 + eps, 0), RX(np.pi / 2 - eps, 1), CZ(0, 1), RX(np.pi / 2 + eps, 1))
    noise_model = _decoherence_noise_model(_get_program_gates(p))
    pnoisy = apply_noise_model(p, noise_model)
    for i in pnoisy:
        if isinstance(i, DefGate):
            pass
        elif isinstance(i, Pragma):
            assert i.command in ["ADD-KRAUS", "READOUT-POVM"]
        elif isinstance(i, Gate):
            assert i.name in NO_NOISE or not i.params



from typing import Union

import numpy as np
from typing import List, Optional, cast
from pyquil.external.rpcq import (
    GateInfo,
    MeasureInfo,
    Supported1QGate,
    Supported2QGate,
    CompilerISA,
    add_qubit,
    add_edge,
)

import networkx as nx


DEFAULT_1Q_GATES = [
    Supported1QGate.I,
    Supported1QGate.RX,
    Supported1QGate.RZ,
    Supported1QGate.MEASURE,
]
DEFAULT_2Q_GATES = [
    Supported2QGate.CZ,
    Supported2QGate.XY,
]


def graph_to_compiler_isa(
    graph: nx.Graph, gates_1q: Optional[List[str]] = None, gates_2q: Optional[List[str]] = None
) -> CompilerISA:
    """
    Generate an ``CompilerISA`` object from a NetworkX graph and list of 1Q and 2Q gates.
    May raise ``GraphGateError`` if the specified gates are not supported.

    :param graph: The graph topology of the device.
    :param gates_1q: A list of 1Q gate names to be made available for all qubits in the device.
    Defaults to ``DEFAULT_1Q_GATES``.
    :param gates_2q: A list of 2Q gate names to be made available for all edges in the device.
    Defaults to ``DEFAULT_2Q_GATES``.
    """
    gates_1q = gates_1q or DEFAULT_1Q_GATES.copy()
    gates_2q = gates_2q or DEFAULT_2Q_GATES.copy()

    device = CompilerISA()

    qubit_gates = []
    for gate in gates_1q:
        qubit_gates.extend(_transform_qubit_operation_to_gates(gate))

    all_qubits = list(range(max(graph.nodes) + 1))
    for i in all_qubits:
        qubit = add_qubit(device, i)
        qubit.gates = qubit_gates
        qubit.dead = i not in graph.nodes

    edge_gates = []
    for gate in gates_2q:
        edge_gates.extend(_transform_edge_operation_to_gates(gate))

    for a, b in graph.edges:
        edge = add_edge(device, a, b)
        edge.gates = edge_gates

    return device


def compiler_isa_to_graph(compiler_isa: CompilerISA) -> nx.Graph:
    """
    Generate an ``nx.Graph`` based on the qubits and edges of any ``CompilerISA``.
    """
    return nx.from_edgelist([int(i) for i in edge.ids] for edge in compiler_isa.edges.values())


class GraphGateError(ValueError):
    """
    Signals an error when creating a ``CompilerISA`` from an ``nx.Graph``.
    This may raise as a consequence of unsupported gates.
    """

    pass


def _make_i_gates() -> List[GateInfo]:
    return [GateInfo(operator=Supported1QGate.I, parameters=[], arguments=["_"])]


def _make_measure_gates() -> List[MeasureInfo]:
    return [
        MeasureInfo(operator=Supported1QGate.MEASURE, qubit="_", target="_"),
        MeasureInfo(operator=Supported1QGate.MEASURE, qubit="_", target=None),
    ]


def _make_rx_gates() -> List[GateInfo]:
    gates = []
    for param in [0.0, np.pi, -np.pi, np.pi / 2, -np.pi / 2]:
        gates.append(GateInfo(operator=Supported1QGate.RX, parameters=[param], arguments=["_"]))
    return gates


def _make_rz_gates() -> List[GateInfo]:
    return [GateInfo(operator=Supported1QGate.RZ, parameters=["theta"], arguments=["_"])]


def _make_wildcard_1q_gates() -> List[GateInfo]:
    return [GateInfo(operator="_", parameters=["_"], arguments=["_"])]


def _transform_qubit_operation_to_gates(
    operation_name: str,
) -> List[Union[GateInfo, MeasureInfo]]:
    if operation_name == Supported1QGate.I:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_i_gates())
    elif operation_name == Supported1QGate.RX:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_rx_gates())
    elif operation_name == Supported1QGate.RZ:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_rz_gates())
    elif operation_name == Supported1QGate.MEASURE:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_measure_gates())
    elif operation_name == Supported1QGate.WILDCARD:
        return cast(List[Union[GateInfo, MeasureInfo]], _make_wildcard_1q_gates())
    else:
        raise GraphGateError("Unsupported graph qubit operation: {}".format(operation_name))


def _make_cz_gates() -> List[GateInfo]:
    return [GateInfo(operator=Supported2QGate.CZ, parameters=[], arguments=["_", "_"])]


def _make_iswap_gates() -> List[GateInfo]:
    return [GateInfo(operator=Supported2QGate.ISWAP, parameters=[], arguments=["_", "_"])]


def _make_cphase_gates() -> List[GateInfo]:
    return [GateInfo(operator=Supported2QGate.CPHASE, parameters=["theta"], arguments=["_", "_"])]


def _make_xy_gates() -> List[GateInfo]:
    return [GateInfo(operator=Supported2QGate.XY, parameters=["theta"], arguments=["_", "_"])]


def _make_wildcard_2q_gates() -> List[GateInfo]:
    return [GateInfo(operator="_", parameters=["_"], arguments=["_", "_"])]


def _transform_edge_operation_to_gates(operation_name: str) -> List[GateInfo]:
    if operation_name == Supported2QGate.CZ:
        return _make_cz_gates()
    elif operation_name == Supported2QGate.ISWAP:
        return _make_iswap_gates()
    elif operation_name == Supported2QGate.CPHASE:
        return _make_cphase_gates()
    elif operation_name == Supported2QGate.XY:
        return _make_xy_gates()
    elif operation_name == Supported2QGate.WILDCARD:
        return _make_wildcard_2q_gates()
    else:
        raise GraphGateError("Unsupported graph edge operation: {}".format(operation_name))
